/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This program is written in GAP-L as developed by Georg Sauthoff at Bielefeld
 * University. GAP-L, the Bellman's Gap Compiler and other language features are
 * used in accordance to the GPL v3+.
 *
 * Bellman's Gap can be found at http://www.gapc.eu
 */

/**
 * This implements the global sequence alignment on simple character sequences.
 * Every character is interpreted as a discrete symbol.
 *
 * @author =  Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
//just some preamble definitions

import align_cost
import twoline_Rope
//import limit_id

input < raw, raw >


type Rope = extern
type twoline_Rope = extern

/*
 * Signature that defines the operations allowed.
 */
signature sig_alignment(alphabet, answer) {
  answer del(<alphabet,void>,answer);
  answer skip_del(<alphabet,void>,answer);
  answer ins(<void,alphabet>,answer);
  answer skip_ins(<void,alphabet>,answer);
  answer rep(<alphabet,alphabet>,answer);
  answer nil(<void,void>);
  choice [answer] h([answer]);
}

/*
 * Algebra section.
 */
algebra enum auto enum ;
algebra count auto count;

algebra alg_strict implements sig_alignment(alphabet = char, answer = double){

	double del(<char left, void>, double in){
		return in + del_cost(left);
	}
	
	double skip_del(<char left, void>, double in){
		return in + skip_del_cost(left);
	}
	
	double ins(<void, char right>, double in){
		return in + ins_cost(right);
	}

	double skip_ins(<void, char right>, double in){
		return in + skip_ins_cost(right);
	}

	double rep(<char left, char right>, double in){
		return in + rep_cost(left, right);
	}
	
	double nil(<void, void>){
		return 0;
	}
	
	choice [double] h([double] l){
		return list(minimum(l));
	}
}

algebra alg_min_gaps implements sig_alignment(alphabet = char, answer = int){

	int del(<char left, void>, int in){
		return in + 1;
	}
	
	int skip_del(<char left, void>, int in){
		return in + 1;
	}
	
	int ins(<void, char right>, int in){
		return in + 1;
	}

	int skip_ins(<void, char right>, int in){
		return in + 1;
	}

	int rep(<char left, char right>, int in){
		return in;
	}
	
	int nil(<void, void>){
		return 0;
	}
	
	choice [int] h([int] l){
		return list(minimum(l));
	}
}

algebra alg_print implements sig_alignment(alphabet = char, answer = twoline_Rope){

	twoline_Rope del(<char left,void>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, left);
		append(out.top, in.top);
		append(out.bottom, '-');
		append(out.bottom, in.bottom);
		return out;
	}

	twoline_Rope skip_del(<char left,void>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, left);
		append(out.top, in.top);
		append(out.bottom, '_');
		append(out.bottom, in.bottom);
		return out;
	}

	twoline_Rope ins(<void,char right>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, '-');
		append(out.top, in.top);
		append(out.bottom, right);
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope skip_ins(<void,char right>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, '_');
		append(out.top, in.top);
		append(out.bottom, right);
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope rep(<char left, char right>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, left);
		append(out.top, in.top);
		append(out.bottom, right);
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope nil(<void,void>){
		twoline_Rope out;
		return out;
	}
	
	choice [twoline_Rope] h([twoline_Rope] l){
		return l;
	}
}

/*
 * Grammar to define what kind of alignments are allowed and which are not.
 * Note that we take the following precautions:
 * - We enforce that at least one replacement happens before we can start
 *   skipping symbols at the end.
 * - We enforce that deletions always are in front of insertions.
 * - We enforce that skip regions in the middle have at least length 4 to
 *   prevent trivial skip regions.
 */
grammar gra_alignment uses sig_alignment(axiom = SKIPDEL_START)
{

	//skip the start
	SKIPDEL_START = skip_del(<CHAR,EMPTY>, SKIPDEL_START) | SKIPINS_START #h;
	
	SKIPINS_START = skip_ins(<EMPTY,CHAR>, SKIPINS_START) |
					rep(<CHAR,CHAR>, ALI) |
					nil(<EMPTY,EMPTY>) #h;

	//align in the middle
	ALI = del(<CHAR,EMPTY>, DEL) | ins(<EMPTY,CHAR>, INS) | rep(<CHAR,CHAR>, ALI) |
			SKIPDEL_MIDDLE | SKIPINS_MIDDLE | SKIPDEL_END #h;
	
	DEL = del(<CHAR,EMPTY>, DEL) | ins(<EMPTY,CHAR>, INS) | rep(<CHAR,CHAR>, ALI) #h;

	INS = ins(<EMPTY,CHAR>, INS) | rep(<CHAR,CHAR>, ALI) #h;
	
	//skip to a next interesting region with at least 4 skips
	SKIPDEL_MIDDLE = skip_del(<CHAR,EMPTY>,
						skip_del(<CHAR,EMPTY>,
						skip_del(<CHAR,EMPTY>,
						skip_del(<CHAR,EMPTY>,
						SKIPDEL_MIDDLE_LOOP)))) #h;
	
	SKIPDEL_MIDDLE_LOOP = skip_del(<CHAR,EMPTY>, SKIPDEL_MIDDLE_LOOP) |
							rep(<CHAR,CHAR>,ALI) #h;
	
	SKIPINS_MIDDLE = skip_ins(<EMPTY,CHAR>,
						skip_ins(<EMPTY,CHAR>,
						skip_ins(<EMPTY,CHAR>,
						skip_ins(<EMPTY,CHAR>,
						SKIPINS_MIDDLE_LOOP)))) #h;

	SKIPINS_MIDDLE_LOOP = skip_ins(<EMPTY,CHAR>, SKIPINS_MIDDLE_LOOP) |
							rep(<CHAR,CHAR>,ALI) #h;

	//skip the end
	SKIPDEL_END = skip_del(<CHAR,EMPTY>, SKIPDEL_END) | SKIPINS_END #h;

	SKIPINS_END = skip_ins(<EMPTY,CHAR>, SKIPINS_END) | nil(<EMPTY,EMPTY>) #h;
}


/*
 * This defines compilable instances of this program.
 * The command line usage (if gapc is installed):
 *
 * gapc -i <instance name> -o local_alignment.cc local_alignment.gap
 *
 */
instance print	= gra_alignment(alg_print);
instance score	= gra_alignment(alg_strict);
instance path	= gra_alignment(alg_strict * alg_min_gaps * alg_print);
instance debug	= gra_alignment(alg_print * alg_strict);
