% TCS Alignment Toolbox
%
% Copyright (C) 2013, 2014
% Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
% AG Theoretical Computer Science
% Centre of Excellence Cognitive Interaction Technology (CITEC)
% University of Bielefeld
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
% 
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

% --- String Edit Distance Example ---

%
% In this example we compare multiple strings in a simple
% StringEditDistance fashion. We also apply a learning scheme to increase
% class separation using metric learning only.

%% Preamble
%
% This includes the necessary classes from the TCSAlignmentToolbox
% .jar-file.


clear java;
clear;
clc;

javaaddpath TCSAlignmentToolbox-1.3.1.jar;
import de.unibi.citec.tcs.alignment.sequence.*;
import de.unibi.citec.tcs.alignment.*;
import de.unibi.citec.tcs.alignment.comparators.*;
import de.unibi.citec.tcs.alignment.wrappers.StringEditDistance;

%% 1.) Set up your data space. In this case we assume to have two
% distinct classes of strings that only consists of 'a's and 'b'. The only
% difference between both classes is, that the members of the second one
% are longer.


strings = javaArray('java.lang.String',6);

% first class with shorter sequences.
strings(1) = java.lang.String('ababababa');
strings(2) = java.lang.String('babababab');
strings(3) = java.lang.String('bababababb');
% second class with longer sequences.
strings(4) = java.lang.String('ababababababbbb');
strings(5) = java.lang.String('bababbababababab');
strings(6) = java.lang.String('aabababaaabbabab');

%% 2.) convert to the TCSAlignmentToolbox sequence format.

sequences = StringEditDistance.toSequences(strings);

% show the first sequence again for testing purposes
sequences(1).toString

%% 3.) Set up the default edit distance scoring scheme.
% Matches cost nothing, everything else costs 1.

alignSpec = StringEditDistance.setUpSpecification(sequences);

%% 4.) Set up your algorithm of choice.

algo = StrictAlignmentFullAlgorithm(alignSpec);

% show one example alignment to see, if our scoring scheme works.

algo.calculateAlignment(sequences(1),sequences(4))

%% 5.) calculate the distance matrix in our data space in parallel.

algo = StrictAlignmentScoreAlgorithm(alignSpec);

parallelEngine = ParallelProcessingEngine(algo,sequences);
parallelEngine.setNumberOfThreads(32);
parallelEngine.calculate();
scoreMat = parallelEngine.getScoreMatrix;

% plot the matrix

figure(1);
imagesc(scoreMat);
colorbar;
title('distance matrix');

% this separates the classes already quite well, but it is not yet optimal.

%% 6.) calculate the derivative.

% we observe, if we look at the derivative inside a class and between
% classes, that more deletions and insertions are used between classes.

algo = StrictAllOptimalAlgorithm(alignSpec);

deriv = StringEditDistance.calculateDerivative(algo,sequences(1),sequences(2));

% plot the matrix of parameter derivatives for an intra-class alignment an
% an inter-class alignment. Inside the class we observe a higher
% replacement derivative (the section (1,1)-(2,2) of the matrix, where the
% diagonal models replacements of equal symbols).
% Between classes we observe more insertions (becuse the second class has
% longer sequences; the section (3,1)-(3,2) of the matrix, where (3,1) is
% an insertion of a and (3,2) is an insertion of b).

figure(2);
imagesc(deriv(1:3,1:3));
set(gca,'XTick',[1,2,3]);
set(gca,'XTickLabel',{'a','b','-'});
set(gca,'YTick',[1,2,3]);
set(gca,'YTickLabel',{'a','b','-'});
colorbar;
title('replacement derivative matrix (intra-class)');

deriv = StringEditDistance.calculateDerivative(algo,sequences(1),sequences(4));

figure(3);
imagesc(deriv(1:3,1:3));
set(gca,'XTick',[1,2,3]);
set(gca,'XTickLabel',{'a','b','-'});
set(gca,'YTick',[1,2,3]);
set(gca,'YTickLabel',{'a','b','-'});
colorbar;
title('replacement derivative matrix (inter-class)');

%% 7.) let's check if that observation holds for the entire data space.

% the last parameter here is the number of threads for parallel
% computation.
derivs = StringEditDistance.calculateDerivatives(algo,sequences,32);

% this methods returns _all_ pairwise derivatives. To get an overview of
% the overall derivative differences we add all intra-class-derivatives and
% subtract all inter-class-derivatives.

deriv = zeros(3,3);

% first class intra-class derivatives.
for i=1:3
    for j = 1:3
        deriv = deriv + squeeze(derivs(i,j,1:3,1:3));
    end
end
% second class intra-class derivatives.
for i=4:6
    for j = 4:6
        deriv = deriv + squeeze(derivs(i,j,1:3,1:3));
    end
end
% first to second class inter-class derivatives.
for i=1:3
    for j = 4:6
        deriv = deriv - squeeze(derivs(i,j,1:3,1:3));
    end
end
% second to first class inter-class derivatives.
for i=4:6
    for j = 1:3
        deriv = deriv - squeeze(derivs(i,j,1:3,1:3));
    end
end

% plot the resulting matrix

figure(4);
imagesc(deriv(1:3,1:3));
set(gca,'XTick',[1,2,3]);
set(gca,'XTickLabel',{'a','b','-'});
set(gca,'YTick',[1,2,3]);
set(gca,'YTickLabel',{'a','b','-'});
colorbar;
title('GLVQ cost function derivative for replacement parameters');

% This resulting matrix actually corresponds to a (very simple) gradient on
% the GLVQ cost function: To improve class separation we have to decrease
% replacement costs (positive matrix values) and have to increase gap costs
% (strongly negative matrix values).

%% 8.) set up new alignment specification according to the
% observation. We use zero replacement costs and maximum (1) gap costs.

% the second parameter is the cost for matches, the second one for
% mismatches and the third one for gaps.
alignSpec = StringEditDistance.setUpSpecification(sequences,0,0,1);

% calculate the distance matrix in our data space again.

algo = StrictAlignmentScoreAlgorithm(alignSpec);

parallelEngine = ParallelProcessingEngine(algo,sequences);
parallelEngine.setNumberOfThreads(32);
parallelEngine.calculate();
scoreMat = parallelEngine.getScoreMatrix;

% plot the matrix

figure(1);
imagesc(scoreMat);
colorbar;
title('distance matrix');

% now we have near perfect class separation.