/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.DerivableComparator;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;

/**
 * This allows parallel processing of derivative calculations.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class ParallelDerivativeEngine {

	public static final int DEFAULTTHREADNUM = 4;

	private final AlignmentDerivativeAlgorithm[] derivativeAlgorithms;
	private int numberOfThreads = DEFAULTTHREADNUM;
	private boolean verbose = true;
	private boolean ignoreExceptions = false;

	/**
	 * This sets up a ParallelProcessingEngine for the given AlignmentAlgorithm
	 * and the given Sequences.
	 *
	 * @param derivativeAlgorithms the derivative algorithms that shall be
	 * exectuted.
	 */
	public ParallelDerivativeEngine(
			AlignmentDerivativeAlgorithm[] derivativeAlgorithms) {
		this.derivativeAlgorithms = derivativeAlgorithms;
	}

	public ParallelDerivativeEngine(
			AlignmentDerivativeAlgorithm[][] derivativeAlgorithms) {
		final ArrayList<AlignmentDerivativeAlgorithm> algos
				= new ArrayList<AlignmentDerivativeAlgorithm>();
		for (int i = 0; i < derivativeAlgorithms.length; i++) {
			if (derivativeAlgorithms[i] == null) {
				continue;
			}
			for (int j = 0; j < derivativeAlgorithms[i].length; j++) {
				if (derivativeAlgorithms[i][j] != null) {
					algos.add(derivativeAlgorithms[i][j]);
				}
			}
		}
		this.derivativeAlgorithms = algos.toArray(new AlignmentDerivativeAlgorithm[0]);
	}

	/**
	 *
	 * @return The AlignmentDerivativeAlgorithm that was given in the
	 * constructor.
	 */
	public AlignmentDerivativeAlgorithm[] getAlgorithms() {
		return derivativeAlgorithms;
	}

	/**
	 * If verbose is set to true the ParallelProcessingEngine will give
	 * information about the current state of the calculation.
	 *
	 * @param verbose true or false.
	 */
	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}

	/**
	 * If verbose is set to true the ParallelProcessingEngine will give
	 * information about the current state of the calculation.
	 *
	 * @return true or false.
	 */
	public boolean isVerbose() {
		return verbose;
	}

	/**
	 * Sets the number of threads that are used. It is recommended to use ~ 2
	 * times the number of CPUs you have at hand.
	 *
	 * This is DEFAULTTHREADNUM per default.
	 *
	 * @param numThreads the number of threads that are used.
	 */
	public void setNumberOfThreads(int numThreads) {
		if (numThreads < 0) {
			throw new RuntimeException(
					"Minimum number of threads was set to a number < 0");
		}
		this.numberOfThreads = numThreads;
	}

	/**
	 * If this is set to true exceptions in single threads are ignored.
	 * Otherwise the ParallelDerivativeEngine stops calculating as soon as one
	 * thread has an Exception. It is false per default.
	 *
	 * @return true if exceptions are ignored.
	 */
	public boolean isIgnoreExceptions() {
		return ignoreExceptions;
	}

	/**
	 * If this is set to true exceptions in single threads are ignored.
	 * Otherwise the ParallelDerivativeEngine stops calculating as soon as one
	 * thread has an Exception. It is false per default.
	 *
	 * @param ignoreExceptions true if exceptions should be ignored.
	 */
	public void setIgnoreExceptions(boolean ignoreExceptions) {
		this.ignoreExceptions = ignoreExceptions;
	}

	/**
	 * Please refer to the corresponding setter method for more information.
	 *
	 * @return the number of threads that are used.
	 */
	public int getNumberOfThreads() {
		return numberOfThreads;
	}

	/**
	 * This calculates the weight derivatives for each of the input
	 * AlignmentDerivativeAlgorithms. The result is returned as a vector for
	 * each of the input algorithms, resulting in a 2-dimensional array.
	 *
	 * @return A 2-dimensional array containing the weight derivatives for each
	 * of the input AlignmentDerivativeAlgorithms.
	 */
	public double[][] calculateWeightDerivatives() {
		final WeightRunnableFactory factory = new WeightRunnableFactory();
		return calculate(factory, double[].class);
	}

	public <R> R[] calculateComparatorDerivatives(final DerivableComparator<?, R> comp, String keyword) {
		final ComparatorRunnableFactory<R> factory = new ComparatorRunnableFactory<>(comp, keyword);
		return calculate(factory, comp.getResultClass());
	}

	/**
	 * This starts the calculation of the pairwise alignment of the input
	 * Sequencs according to the given algorithm.
	 */
	private <R> R[] calculate(RunnableFactory<R> factory, Class<R> resultClass) {
		ArrayDeque<ThreadMessage> messageQueue = new ArrayDeque<ThreadMessage>();
		final Counter counter = new Counter();
		final int N = derivativeAlgorithms.length;

		final R[] outMatrix = AbstractGapAlignmentAlgorithm.createGenericArray(N, resultClass);
		final ReentrantLock matrixLock = new ReentrantLock();
		final ConcurrentHashMap<Integer, Thread> threads
				= new ConcurrentHashMap<Integer, Thread>();
		for (int i = 0; i < N; i++) {

			//check if we reached the maximum.
			while (threads.size() == numberOfThreads) {
				//If we reached the maximum, join the first thread in the Thread Map.
				try {
					final Thread randomThread = threads.values().iterator().next();
					randomThread.join();
				} catch (InterruptedException ex) {
					throw new RuntimeException(ex);
				}
				//check if threads sent messages.
				handleMessages(messageQueue, matrixLock, N, counter, threads);
			}
			//otherwise create a new thread.
			final Thread newThread = new Thread(factory.create(
					derivativeAlgorithms[i],
					matrixLock, outMatrix, i,
					messageQueue));
			threads.put(i, newThread);
			newThread.start();

		}
		//when we started all calculations wait for them to finish.
		final Iterator<Thread> iterator = threads.values().iterator();
		while (iterator.hasNext()) {
			final Thread thread = iterator.next();
			try {
				thread.join();
			} catch (InterruptedException ex) {
				throw new RuntimeException(ex);
			}
			handleMessages(messageQueue, matrixLock, N, counter, threads);
		}
		handleMessages(messageQueue, matrixLock, N, counter, threads);
		//then set the result matrix.
		if (verbose) {
			System.out.println("Finished calculation!");
			System.out.flush();
		}
		return outMatrix;
	}

	//sent a message for verbose calculation every 5 %.
	public static final int SHOW_STEPS = 5;

	private void handleMessages(ArrayDeque<ThreadMessage> messageQueue,
			ReentrantLock matrixLock, int N, Counter counter,
			ConcurrentHashMap<Integer, Thread> threads) {
		ArrayDeque<ThreadMessage> msgQueueCopy = null;
		matrixLock.lock();
		try {
			if (!messageQueue.isEmpty()) {
				msgQueueCopy = new ArrayDeque<ThreadMessage>(messageQueue);
				messageQueue.clear();
			} else {
				msgQueueCopy = null;
			}
		} finally {
			matrixLock.unlock();
		}
		if (msgQueueCopy != null) {
			while (!msgQueueCopy.isEmpty()) {
				final ThreadMessage msg = msgQueueCopy.poll();
				if (msg.getText().endsWith(" is finished!")) {
					//if a thread is finished, remove it from the thread map.
					final Thread finishedThread = threads.
							remove(msg.getOrigin().threadIdx);
					try {
						//let the thread join pro forma.
						finishedThread.join();
					} catch (InterruptedException ex) {
						throw new RuntimeException(ex);
					}
					//if we want to have additional information, look for the current count.
					if (verbose) {
						final double oldFinishedPartition
								= (double) counter.getValue() / (double) N;
						final int oldStep = ((int) (oldFinishedPartition * 100))
								/ SHOW_STEPS;
						counter.inc();
						final double newFinishedPartition
								= (double) counter.getValue() / (double) N;
						final int newStep = ((int) (newFinishedPartition * 100))
								/ SHOW_STEPS;
						if (newStep > oldStep) {
							System.out.println(
									"Finished "
									+ (int) Math.round(newFinishedPartition * 100)
									+ "% of calculation.");
							System.out.flush();
						}
					}
				} else {
					if (verbose) {
						System.out.println(
								msg.getOrigin().threadIdx + " sent a message: "
								+ msg.getText());
						System.out.flush();
					}
					if (msg instanceof ExceptionThreadMessage) {
						//let the thread join pro forma.
						final Thread exceptionThread = threads.
								remove(msg.getOrigin().threadIdx);
						try {
							exceptionThread.join();
						} catch (InterruptedException ex) {
							throw new RuntimeException(ex);
						}
						if(!ignoreExceptions){
							//if we do not want to ignore exceptions, we stop
							//all calculations and rethrow the exception.
							for(final Thread thread : threads.values()){
								thread.interrupt();
							}
							throw ((ExceptionThreadMessage) msg).getException();
						}
					}
				}
			}
		}
	}

	private static class WeightRunnableFactory implements RunnableFactory<double[]> {

		@Override
		public CalculationRunnable<double[]> create(AlignmentDerivativeAlgorithm algorithm,
				ReentrantLock matrixLock, double[][] resultMatrix, int id,
				ArrayDeque<ThreadMessage> msgQueue) {
			return new WeightRunnable(algorithm, matrixLock, resultMatrix, id, msgQueue);
		}

	}

	private static class ComparatorRunnableFactory<Y> implements RunnableFactory<Y> {

		private final DerivableComparator<?, Y> comparator;
		private final String keyword;

		public ComparatorRunnableFactory(DerivableComparator<?, Y> comparator,
				String keyword) {
			this.comparator = comparator;
			this.keyword = keyword;
		}

		@Override
		public CalculationRunnable<Y> create(AlignmentDerivativeAlgorithm algorithm,
				ReentrantLock matrixLock, Y[] resultVector, int id,
				ArrayDeque<ThreadMessage> msgQueue) {
			return new ComparatorRunnable<>(comparator, keyword, algorithm,
					matrixLock, resultVector, id, msgQueue);
		}

	}

	private static interface RunnableFactory<R> {

		public CalculationRunnable<R> create(AlignmentDerivativeAlgorithm algorithm,
				ReentrantLock matrixLock,
				R[] resultMatrix, int id, ArrayDeque<ThreadMessage> msgQueue);

	}

	private static class WeightRunnable extends CalculationRunnable<double[]> {

		public WeightRunnable(AlignmentDerivativeAlgorithm algorithm, ReentrantLock matrixLock,
				double[][] resultMatrix, int id, ArrayDeque<ThreadMessage> msgQueue) {
			super(algorithm, matrixLock, resultMatrix, id, msgQueue);
		}

		@Override
		public double[] calculate() {
			return super.algorithm.calculateWeightDerivative();
		}
	}

	private static class ComparatorRunnable<Y> extends CalculationRunnable<Y> {

		private final DerivableComparator<?, Y> comparator;
		private final String keyword;

		public ComparatorRunnable(DerivableComparator<?, Y> comparator,
				String keyword,
				AlignmentDerivativeAlgorithm algorithm,
				ReentrantLock matrixLock, Y[] resultVector, int id,
				ArrayDeque<ThreadMessage> msgQueue) {
			super(algorithm, matrixLock, resultVector, id, msgQueue);
			this.comparator = comparator;
			this.keyword = keyword;
		}

		@Override
		public Y calculate() {
			return super.algorithm.calculateParameterDerivative(comparator, keyword);
		}

	}

	private static abstract class CalculationRunnable<R> implements Runnable {

		final AlignmentDerivativeAlgorithm algorithm;
		final ReentrantLock matrixLock;
		final R[] resultVector;
		final int threadIdx;
		final ArrayDeque<ThreadMessage> msgQueue;

		public CalculationRunnable(AlignmentDerivativeAlgorithm algorithm, ReentrantLock matrixLock,
				R[] resultVector, int threadIdx, ArrayDeque<ThreadMessage> msgQueue) {
			this.algorithm = algorithm;
			this.matrixLock = matrixLock;
			this.resultVector = resultVector;
			this.threadIdx = threadIdx;
			this.msgQueue = msgQueue;
		}

		public void run() {
			final R result;
			try {
				result = calculate();
			} catch (RuntimeException ex) {
				//if we had an exception, signify the main thread.
				try {
					matrixLock.lock();
					msgQueue.add(new ExceptionThreadMessage(ex, this));
				} finally {
					matrixLock.unlock();
				}
				//and end the calculation.
				return;
			}
			try {
				matrixLock.lock();
				resultVector[threadIdx] = result;
				msgQueue.add(new ThreadMessage(this, " is finished!"));
			} finally {
				matrixLock.unlock();
			}
		}

		public abstract R calculate();
	}

	private static class ThreadMessage {

		private final CalculationRunnable origin;
		private final String text;

		public ThreadMessage(CalculationRunnable origin, String text) {
			this.origin = origin;
			this.text = text;
		}

		public CalculationRunnable getOrigin() {
			return origin;
		}

		public String getText() {
			return text;
		}
	}

	private static class ExceptionThreadMessage extends ThreadMessage {

		private final RuntimeException ex;

		public ExceptionThreadMessage(RuntimeException ex,
				CalculationRunnable origin) {
			super(origin, ex.getMessage());
			this.ex = ex;
		}

		public RuntimeException getException() {
			return ex;
		}

	}
}
