/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.sequence.Sequence;
import java.util.ArrayDeque;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;

/**
 * This allows parallel processing of alignment calculations.
 *
 * @param <R> The result type of the alignment algorithm.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class ParallelProcessingEngine<R> {

	public static final int DEFAULTTHREADNUM = 4;

	private final AlignmentAlgorithm< R> algorithm;
	private final Sequence[] sequences;
	private final ReentrantLock finishedLock = new ReentrantLock();
	private R[][] resultMatrix;
	private boolean finished = false;
	private int numberOfThreads = DEFAULTTHREADNUM;
	private boolean[][] doCalculation;
	private boolean symmetric;
	private boolean verbose = true;
	private boolean ignoreExceptions = false;

	/**
	 * This sets up a ParallelProcessingEngine for the given AlignmentAlgorithm
	 * and the given Sequences.
	 *
	 * @param algorithm an AlignmentAlgorithm of your choice.
	 * @param sequences an array of Sequences. Per default the
	 * ParallelProcessingEngine will calculate all pairwise alignments between
	 * them.
	 */
	public ParallelProcessingEngine(
			AlignmentAlgorithm< R> algorithm, Sequence[] sequences) {
		this.algorithm = algorithm;
		this.sequences = sequences;
	}

	/**
	 *
	 * @return The AlignmentAlgorithm that was given in the constructor.
	 */
	public AlignmentAlgorithm< R> getAlgorithm() {
		return algorithm;
	}

	/**
	 *
	 * @return The Sequence instances that were given in the constructor.
	 */
	public Sequence[] getSequences() {
		return sequences;
	}

	/**
	 * If verbose is set to true the ParallelProcessingEngine will give
	 * information about the current state of the calculation.
	 *
	 * @param verbose true or false.
	 */
	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}

	/**
	 * If verbose is set to true the ParallelProcessingEngine will give
	 * information about the current state of the calculation.
	 *
	 * @return true or false.
	 */
	public boolean isVerbose() {
		return verbose;
	}

	
	/**
	 * If this is set to true exceptions in single threads are ignored.
	 * Otherwise the ParallelDerivativeEngine stops calculating as soon as one
	 * thread has an Exception. It is false per default.
	 *
	 * @return true if exceptions are ignored.
	 */
	public boolean isIgnoreExceptions() {
		return ignoreExceptions;
	}

	/**
	 * If this is set to true exceptions in single threads are ignored.
	 * Otherwise the ParallelDerivativeEngine stops calculating as soon as one
	 * thread has an Exception. It is false per default.
	 *
	 * @param ignoreExceptions true if exceptions should be ignored.
	 */
	public void setIgnoreExceptions(boolean ignoreExceptions) {
		this.ignoreExceptions = ignoreExceptions;
	}
	
	/**
	 * This enables a user to specifically set which pairwise similarity
	 * calculations should be done and which should not be done.
	 *
	 * Say you gave S sequences in the constructor. Than the input of this
	 * setter has to be a S x S - matrix.
	 *
	 * The ParallelProcessingEngine will calculate the alignment between the
	 * sequences i and j if and only if you set the matrix entry (i,j) to true.
	 *
	 * @param doCalculation a S x S - matrix that specifies which pairwise
	 * alignment calculations should be done.
	 */
	public void setDoCalculation(boolean[][] doCalculation) {
		//check size
		if (doCalculation.length != sequences.length) {
			throw new RuntimeException("Tried to set a DoCalculation matrix with wrong dimensions!");
		}
		for (int i = 0; i < sequences.length; i++) {
			if (doCalculation[i].length != sequences.length) {
				throw new RuntimeException(
						"Tried to set a DoCalculation matrix with wrong dimensions!");
			}
		}
		this.doCalculation = doCalculation;
	}

	/**
	 * Please refer to the corresponding setter method for more information.
	 *
	 * @return null per default or the matrix you set using the setter method.
	 */
	public boolean[][] getDoCalculation() {
		return doCalculation;
	}

	/**
	 * Sets the number of threads that are used. It is recommended to use ~ 2
	 * times the number of CPUs you have at hand.
	 *
	 * This is DEFAULTTHREADNUM per default.
	 *
	 * @param numThreads the number of threads that are used.
	 */
	public void setNumberOfThreads(int numThreads) {
		if (numThreads < 0) {
			throw new RuntimeException(
					"Minimum number of threads was set to a number < 0");
		}
		this.numberOfThreads = numThreads;
	}

	/**
	 * Please refer to the corresponding setter method for more information.
	 *
	 * @return the number of threads that are used.
	 */
	public int getNumberOfThreads() {
		return numberOfThreads;
	}

	/**
	 * If this is set to true it is assumed that the desired result of the
	 * alignment is symmetric, meaning that the alignment between the sequences
	 * i and j is equal to the alignment between j and i. If this is the case
	 * the number of alignment calculations can be cut in half by copying the
	 * result from (i,j) to the (j,i) cell.
	 *
	 * @param symmetric true or false.
	 */
	public void setSymmetric(boolean symmetric) {
		this.symmetric = symmetric;
	}

	/**
	 * Please refer to the corresponding setter method for more information.
	 *
	 * @return true or false.
	 */
	public boolean isSymmetric() {
		return symmetric;
	}

	/**
	 * This returns the matrix of results if the calculation is done or null if
	 * it has not finished yet.
	 *
	 * @return the matrix of results if the calculation is done or null if it
	 * has not finished yet.
	 */
	public R[][] getResultMatrix() {
		try {
			finishedLock.lock();
			return resultMatrix;
		} finally {
			finishedLock.unlock();
		}
	}

	/**
	 * This method will only work if the Result class is Double.
	 *
	 * @return the primitive double matrix of the result.
	 */
	public double[][] getScoreMatrix() {
		try {
			finishedLock.lock();
			if (!algorithm.getResultClass().getName().endsWith("Double")) {
				throw new RuntimeException(
						"Attempted to retrieve a primitive score matrix but the Algorithm does not return a primitive double but "
						+ algorithm.getResultClass().getName());
			}
			if (resultMatrix == null) {
				return null;
			}
			final double[][] scoreMatrix = new double[resultMatrix.length][resultMatrix.length];
			for (int i = 0; i < scoreMatrix.length; i++) {
				for (int j = 0; j < scoreMatrix.length; j++) {
					if (resultMatrix[i][j] == null) {
						scoreMatrix[i][j] = Double.NaN;
					} else {
						scoreMatrix[i][j] = (Double) resultMatrix[i][j];
					}
				}
			}
			return scoreMatrix;
		} finally {
			finishedLock.unlock();
		}
	}

	/**
	 *
	 * @return true if the calculation is finished or false if it is not.
	 */
	public boolean isFinished() {
		try {
			finishedLock.lock();
			return finished;
		} finally {
			finishedLock.unlock();
		}
	}

	/**
	 * This starts the calculation of the pairwise alignment of the input
	 * Sequencs according to the given algorithm.
	 */
	public void calculate() {
		ArrayDeque<ThreadMessage<R>> messageQueue = new ArrayDeque<ThreadMessage<R>>();
		final Counter counter = new Counter();
		int maxNum = 0;
		if (verbose) {
			if (doCalculation == null) {
				maxNum = sequences.length * sequences.length;
				if (symmetric) {
					maxNum /= 2;
				}
			} else {
				for (int i = 0; i < sequences.length; i++) {
					for (int j = 0; j < sequences.length; j++) {
						if (doCalculation[i][j]) {
							maxNum++;
						}
					}
				}
			}
		}
		final R[][] tempMatrix = AbstractGapAlignmentAlgorithm.createGenericMatrix(
				sequences.length, sequences.length, algorithm.getResultClass());
		final ReentrantLock matrixLock = new ReentrantLock();
		final ConcurrentHashMap<ThreadID, Thread> threads
				= new ConcurrentHashMap<ThreadID, Thread>();
		for (int i = 0; i < sequences.length; i++) {
			for (int j = 0; j < sequences.length; j++) {
				//check if we have to do this calculation.
				if (doCalculation != null && !doCalculation[i][j]) {
					continue;
				}
				if (symmetric && j < i) {
					continue;
				}
				//check if we reached the maximum.
				while (threads.size() == numberOfThreads) {
					//If we reached the maximum, join the first thread in the Thread Map.
					try {
						final Thread randomThread = threads.values().iterator().next();
						randomThread.join();
					} catch (InterruptedException ex) {
						throw new RuntimeException(ex);
					}
					//handle messages of threads.
					handleMessages(messageQueue, matrixLock, maxNum, counter,
							threads);
				}
				//otherwise create a new thread.
				final ThreadID threadID = new ThreadID(i, j);
				final Thread newThread = new Thread(new CalculationRunnable(
						algorithm, sequences[i], sequences[j],
						matrixLock, tempMatrix, threadID,
						verbose, messageQueue));
				threads.put(threadID, newThread);
				newThread.start();
			}
		}
		//when we started all calculations wait for them to finish.
		final Iterator<Thread> iterator = threads.values().iterator();
		while (iterator.hasNext()) {
			final Thread thread = iterator.next();
			try {
				thread.join();
			} catch (InterruptedException ex) {
				throw new RuntimeException(ex);
			}
			//handle messages of threads.
			handleMessages(messageQueue, matrixLock, maxNum, counter,
					threads);
		}
		//handle messages of threads.
		handleMessages(messageQueue, matrixLock, maxNum, counter,
				threads);
		//if we are in symmetric mode we also adjust the other half of the matrix.
		if (symmetric) {
			for (int i = 0; i < sequences.length; i++) {
				for (int j = i + 1; j < sequences.length; j++) {
					tempMatrix[j][i] = tempMatrix[i][j];
				}
			}
		}
		//then set the result matrix.
		try {
			finishedLock.lock();
			resultMatrix = tempMatrix;
			finished = true;
		} finally {
			finishedLock.unlock();
		}
		if (verbose) {
			System.out.println("Finished calculation!");
		}
	}

	//sent a message for verbose calculation every 5 %.
	public static final int SHOW_STEPS = 5;

	private void handleMessages(ArrayDeque<ThreadMessage<R>> messageQueue,
			ReentrantLock matrixLock, int maxNum, Counter counter,
			ConcurrentHashMap<ThreadID, Thread> threads) {

		//check if threads sent messages.
		ArrayDeque<ThreadMessage<R>> msgQueueCopy = null;
		matrixLock.lock();
		try {
			if (!messageQueue.isEmpty()) {
				msgQueueCopy = new ArrayDeque<ThreadMessage<R>>(messageQueue);
				messageQueue.clear();
			} else {
				msgQueueCopy = null;
			}
		} finally {
			matrixLock.unlock();
		}
		if (msgQueueCopy != null) {
			while (!msgQueueCopy.isEmpty()) {
				final ThreadMessage msg = msgQueueCopy.poll();
				if (msg.getText().endsWith(" is finished!")) {
					//if a thread is finished, remove it from the thread map.
					final Thread finishedThread = threads.
							remove(msg.getOrigin().getId());
					try {
						//let the thread join pro forma.
						finishedThread.join();
					} catch (InterruptedException ex) {
						throw new RuntimeException(ex);
					}
					//if we want to have additional information, look for the current count.
					if (verbose) {
						final double oldFinishedPartition = (double) counter.getValue()
								/ (double) maxNum;
						final int oldStep = ((int) (oldFinishedPartition * 100))
								/ SHOW_STEPS;
						counter.inc();
						final double newFinishedPartition = (double) counter.getValue()
								/ (double) maxNum;
						final int newStep = ((int) (newFinishedPartition * 100))
								/ SHOW_STEPS;
						if (newStep > oldStep) {
							System.out.println(
									"Finished "
									+ (int) Math.round(newFinishedPartition * 100)
									+ "% of calculation.");
						}
					}
				} else {
					if (verbose) {
						System.out.println(
								msg.getOrigin().getId() + " sent a message: "
								+ msg.getText());
					}
					if (msg instanceof ExceptionThreadMessage) {
						//let the thread join pro forma.
						final Thread exceptionThread = threads.
								remove(msg.getOrigin().getId());
						try {
							exceptionThread.join();
						} catch (InterruptedException ex) {
							throw new RuntimeException(ex);
						}
						if(!ignoreExceptions){
							//if we do not want to ignore exceptions, we stop
							//all calculations and rethrow the exception.
							for(final Thread thread : threads.values()){
								thread.interrupt();
							}
							throw ((ExceptionThreadMessage) msg).getException();
						}
					}
				}

			}
		}
	}

	private static class ThreadID {

		private final int i;
		private final int j;

		public ThreadID(int i, int j) {
			this.i = i;
			this.j = j;
		}

		public int getI() {
			return i;
		}

		public int getJ() {
			return j;
		}

		@Override
		public int hashCode() {
			int hash = 7;
			hash = 41 * hash + this.i;
			hash = 41 * hash + this.j;
			return hash;
		}

		@Override
		public boolean equals(Object obj) {
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			final ThreadID other = (ThreadID) obj;
			if (this.i != other.i) {
				return false;
			}
			if (this.j != other.j) {
				return false;
			}
			return true;
		}

		@Override
		public String toString() {
			return "ThreadID{" + "i=" + i + ", j=" + j + '}';
		}
	}

	private static class CalculationRunnable<R> implements Runnable {

		private final AlignmentAlgorithm< R> algorithm;
		private final Sequence seq1;
		private final Sequence seq2;
		private final ReentrantLock matrixLock;
		private final R[][] resultMatrix;
		private final ThreadID id;
		private final boolean verbose;
		private final ArrayDeque<ThreadMessage> msgQueue;

		public CalculationRunnable(AlignmentAlgorithm< R> algorithm,
				Sequence seq1, Sequence seq2,
				ReentrantLock matrixLock, R[][] resultMatrix,
				ThreadID id,
				boolean verbose, ArrayDeque<ThreadMessage> msgQueue) {
			this.algorithm = algorithm;
			this.seq1 = seq1;
			this.seq2 = seq2;
			this.matrixLock = matrixLock;
			this.resultMatrix = resultMatrix;
			this.id = id;
			this.verbose = verbose;
			this.msgQueue = msgQueue;
		}

		public ThreadID getId() {
			return id;
		}

		public void run() {
			final R result;
			try {
				result = algorithm.calculateAlignment(seq1, seq2);
			} catch (RuntimeException ex) {
				//if we had an exception, signify the main thread.
				try {
					matrixLock.lock();
					msgQueue.add(new ExceptionThreadMessage(ex, this));
				} finally {
					matrixLock.unlock();
				}
				//and end the calculation.
				return;
			}
			try {
				matrixLock.lock();
				resultMatrix[id.getI()][id.getJ()] = result;
				msgQueue.add(new ThreadMessage(this, " is finished!"));
			} finally {
				matrixLock.unlock();
			}
		}
	}

	private static class ThreadMessage<R> {

		private final CalculationRunnable<R> origin;
		private final String text;

		public ThreadMessage(CalculationRunnable<R> origin, String text) {
			this.origin = origin;
			this.text = text;
		}

		public CalculationRunnable<R> getOrigin() {
			return origin;
		}

		public String getText() {
			return text;
		}
	}

	private static class ExceptionThreadMessage<R> extends ThreadMessage<R> {

		private final RuntimeException ex;

		public ExceptionThreadMessage(RuntimeException ex, CalculationRunnable<R> origin) {
			super(origin, ex.getMessage());
			this.ex = ex;
		}

		public RuntimeException getException() {
			return ex;
		}
	}
}
