/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.sequence.Sequence;

/**
 * This calculates the soft (global) alignment of two sequences similarly to the
 * Needleman-Wunsch-Algorithm. The whole trick of this
 * SoftAlignmentScoreAlgorithm-Algorithm is to replace the strict minimum in the
 * cost function by a softmin approximation. Thereby we give the alignment a
 * "probabilistic touch": Even suboptimal alignments are considered by the
 * softmin approach, but less than the optimal alignment.
 *
 * This algorithm returns just the accumulated, normalized score. In each step
 * the score is 1 at max and in the worst case m+n steps are needed. Thus the
 * normalization is done by dividing by m+n.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class SoftAlignmentScoreAlgorithm
        extends AbstractGapAlignmentAlgorithm<SoftAlignmentScoreAlgorithm.ScoreEntry, Double> {

    public static final double DEFAULTBETA = 1;
    private double beta = DEFAULTBETA;

    /**
     * This sets up an AlignmentAlgorithm instance according to the given
     * specification.
     *
     * @param alignmentSpecification an AlignmentSpecification.
     */
    public SoftAlignmentScoreAlgorithm(AlignmentSpecification alignmentSpecification) {
        super(alignmentSpecification, SoftAlignmentScoreAlgorithm.ScoreEntry.class, Double.class);
    }

    /**
     *
     * @param beta The parameter defining the "softness" of the alignment. For
     * beta towards infinity this alignment becomes closer to the strict
     * alignment. For beta = 0 all possible alignments are equally considered
     * and softmin returns the average. Please note that a low beta value might
     * lead to a very rough approximation and that for higher sequence lengths
     * beta has to be higher, too.
     */
    public void setBeta(double beta) {
        this.beta = beta;
    }

    /**
     *
     * @return The parameter defining the "softness" of the alignment. For beta
     * towards infinity this alignment becomes closer to the strict alignment.
     * For beta = 0 all possible alignments are equally considered and softmin
     * returns the average. Please note that a low beta value might lead to a
     * very rough approximation and that for higher sequence lengths beta has to
     * be higher, too.
     */
    public double getBeta() {
        return beta;
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public ScoreEntry createInitial() {
        return new ScoreEntry(0);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public ScoreEntry createDelInitial(ScoreEntry delOld, int i, double delLocal) {
        return new ScoreEntry(delOld.getScore() + delLocal);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public ScoreEntry createInsInitial(ScoreEntry insOld, int j, double insLocal) {
        return new ScoreEntry(insOld.getScore() + insLocal);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public ScoreEntry createNewEntry(ScoreEntry delOld, ScoreEntry insOld,
            ScoreEntry repOld,
            int i, int j,
            double delLocal, double insLocal, double repLocal) {
        final double delTotal = delOld.getScore() + delLocal;
        final double insTotal = insOld.getScore() + insLocal;
        final double repTotal = repOld.getScore() + repLocal;
        return new ScoreEntry(softmin(beta, repTotal, delTotal, insTotal));
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public Double transformToResult(ScoreEntry[][] alignmentMatrix,
            Sequence a, Sequence b) {
        final double rawScore = alignmentMatrix[a.getNodes().size()][b.getNodes().size()].
                getScore();
        /*
         * In each step of the alignment we had a distance of 1 at most. Thus we
         * can normalize the score to the range [0,1] if we just divide the raw
         * score by the worst case, which would be to delete a entirely and
         * insert b entirely => worst case score = m+n.
         */
        return rawScore / ((double) (a.getNodes().size() + b.getNodes().size()));
    }

    /**
     * This means that softmin arguments with a weight smaller than 10^-3 will
     * be disregarded. This is a valid approximation because we are garantueed
     * to have at least one weight = 1 which will dominate all very small
     * weights.
     */
    public static final double GLOBALAPPROXTHRESHOLD = 3 * Math.log(10);

    /**
     * This calculates softmin for the given input and the given value of beta.
     *
     * Please note that this approximation can not be 100% correct for very high
     * values of beta or input arguments. However this implementation does a
     * numeric correction such that e.g. the softmin of 100 and 101 can be
     * correctly calculated. For large differences between input arguments (e.g.
     * 100 and 110) however the implementation will most likely return exactly
     * 100, even though that is an approximation.
     *
     * @param beta Softmin equals min for beta towards infinity. For beta = 0
     * softmin returns the average of the given input arguments. Thus beta
     * regulates the "softness" of the softmin approximation.
     * @param args the input arguments.
     *
     * @return a soft approximation of the minimum of all input arguments.
     */
    public static double softmin(double beta, double... args) {
        if (args.length == 0) {
            return 0;
        }
        /*
         * first get the minimum and substract it from all inputs to prevent
         * numeric trouble. This is allowed because softmin returns the same
         * value if we substract a constant.
         */
        double min = args[0];
        int minIdx = 0;
        for (int i = 1; i < args.length; i++) {
            if (args[i] < min) {
                min = args[i];
                minIdx = i;
            }
        }
        /**
         * Then calculate actual softmin.
         */
        double normalization = 0;
        double weightedSum = 0;
        double weight;
        double exponent;
        for (int i = 0; i < args.length; i++) {
            if (i == minIdx) {
                //if this is the minimum then the result is clear.
                normalization += 1;
                weightedSum += args[i];
            } else {
                exponent = beta * (args[i] - min);
                /*
                 * as we are garantueed to have one weight = 1 as baseline we
                 * can approximate very small weights with zero as they will be
                 * dominated by the minimum. This should smooth out some
                 * numerical issues.
                 */
                if (exponent < GLOBALAPPROXTHRESHOLD) {
                    weight = Math.exp(-exponent);
                    normalization += weight;
                    weightedSum += weight * args[i];
                }
            }
        }
        return weightedSum / normalization;
    }

    public static class ScoreEntry implements AlignmentMatrixEntry {

        private final double totalScore;

        public ScoreEntry(double totalScore) {
            this.totalScore = totalScore;
        }

        public double getScore() {
            return totalScore;
        }
    }
}
