/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.ValueType;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Objects;
import java.util.zip.Deflater;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import net.jpountz.lz4.LZ4Compressor;
import net.jpountz.lz4.LZ4Factory;

/**
 * This compares strings using normalized compression distance.
 *
 * For gaps and skips this comparison method is extended by the
 * ComparisonBasedSkipExtendedComparator.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class NCDComparator
		extends ComparisonBasedSkipExtendedComparator<StringValue> {

	private CompressorType compressor = CompressorType.LZ4;

	public NCDComparator() {
		super(new StringValue(""), new StringValue(""),
				new StringValue(""), new StringValue(""));
	}

	public void setCompressor(CompressorType compressor) {
		this.compressor = compressor;
	}

	public CompressorType getCompressor() {
		return compressor;
	}

	/**
	 * {@inheritDoc }
	 */
	public double compare(StringValue a, StringValue b) {
		if (a == null && b == null) {
			return 0;
		}
		if (a == null && b != null) {
			return insert(b);
		}
		if (a != null && b == null) {
			return delete(a);
		}
		return calculateNCD(a.getString(), b.getString());
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ValueType getType() {
		return ValueType.STRING;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int hashCode() {
		int hash = 7;
		hash = 37 * hash + Objects.hashCode(this.compressor);
		hash = 37 * hash + super.hashCode();
		return hash;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final NCDComparator other = (NCDComparator) obj;
		if (this.compressor != other.compressor) {
			return false;
		}
		if (!super.equals(obj)) {
			return false;
		}
		return true;
	}

	/**
	 * Returns the normalized compression distance between two input strings
	 * according to the compressor currently set.
	 *
	 * The normalized compression distance is based on the size of the
	 * compressed strings. Let x and y be the input strings and C(s) the
	 * <strong>size</strong> of the compression of string s.
	 *
	 * Then the NCD value between x and y is given by
	 *
	 * NCD(x,y) := (C(xy) - min{C(x), C(y)}) / max{C(x), C(y)}
	 *
	 * This leads to values of NCD(x,y) between 0 and 1, where 0 means that x
	 * and y are equal and 1 means that x and y are entirely different.
	 *
	 * This description of NCD is based on the article posted by Marc Dommers
	 * here:
	 *
	 * http://www.c-sharpcorner.com/uploadfile/acinonyx72/calculating-the-normalized-compression-distance-between-two-strings/
	 *
	 * @param x first input string.
	 * @param y second input string.
	 * @return the NCD value between both strings.
	 */
	public double calculateNCD(final String x, final String y) {
		try {
			//calculate compressed size for the single inputs ...
			final double cX = compressor.calculateCompressedSize(x);
			final double cY = compressor.calculateCompressedSize(y);
			//... then for the concatenated input;
			final double cXY = compressor.calculateCompressedSize(x + y);
			//then calculate and return NCD
			return Math.max(0, (cXY - Math.min(cX, cY)) / Math.max(cX, cY));
		} catch (IOException ex) {
			throw new RuntimeException(ex);
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double comparisonDerivative(int paramIdx, StringValue a, StringValue b) {
		//this has no further parameters.
		return 0;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public SparseLocalDerivative sparseComparisonDerivative(StringValue a, StringValue b) {
		//this has no further parameters.
		return new SparseLocalDerivative.EmptySparseLocalEntry();
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int getNumberOfComparisonParameters() {
		//this has no further parameters.
		return 0;
	}

	/**
	 * This class contains all possible CompressorTypes for this NCD
	 * implementation.
	 *
	 * @author bpaassen
	 */
	public static enum CompressorType {

		ZLIB, ZIP, LZ4;

		/**
		 * Calculates the compressed size of the given input string in bytes.
		 *
		 * @param input a string.
		 * @return the compressed size of the given input string in bytes.
		 */
		public int calculateCompressedSize(final String input) throws IOException {
			if (input == null) {
				return -1;
			}
			switch (this) {
				case ZLIB: {
					// Encode a String into bytes
					final byte[] inputBytes = input.getBytes("UTF-8");
					// Compress the bytes
					final byte[] output = new byte[100];
					final Deflater compresser = new Deflater(Deflater.BEST_SPEED);
					compresser.setInput(inputBytes);
					compresser.finish();
					final int compressedSize = compresser.deflate(output);
					compresser.end();
					return compressedSize;
				}
				case ZIP: {
					final ByteArrayOutputStream out = new ByteArrayOutputStream();
					final ZipOutputStream zip = new ZipOutputStream(out);
					final ZipEntry zipEntry = new ZipEntry("default");
					zip.putNextEntry(zipEntry);
					zip.write(input.getBytes("UTF-8"));
					zip.closeEntry();
					final int compressedSize = (int) zipEntry.getCompressedSize();
					zip.close();
					return compressedSize;
				}
				case LZ4: {
					final byte[] data = input.getBytes("UTF-8");
					final int decompressedLength = data.length;
					final LZ4Factory factory = LZ4Factory.fastestInstance();
					// compress data
					final LZ4Compressor compressor = factory.fastCompressor();
					final int maxCompressedLength
							= compressor.maxCompressedLength(decompressedLength);
					final byte[] compressed = new byte[maxCompressedLength];
					final int compressedLength = compressor.compress(data, 0,
							decompressedLength, compressed, 0, maxCompressedLength);
					return compressedLength;
				}
			}
			return -1;

		}
	}
}
