/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.sequence;

import java.util.Arrays;

/**
 * This value is reserved for numeric vectors, represented internally as
 * doubles.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class VectorialValue extends AbstractValue {

    private double[] vector;

    public VectorialValue() {
        super(ValueType.VECTOR);
    }

    public VectorialValue(double[] vector) {
        super(ValueType.VECTOR);
        this.vector = vector;
    }

    public VectorialValue(Object convertibleObject) {
        super(ValueType.VECTOR);
        this.vector = mapToDoubleVector(convertibleObject);
    }

    /**
     * Returns the actual vector.
     *
     * @return the actual vector.
     */
    public double[] getVector() {
        return vector;
    }

    /**
     * Sets the actual vector.
     *
     * @param vector the actual vector.
     */
    public void setVector(double[] vector) {
        this.vector = vector;
    }

    /**
     * Sets the actual vector.
     *
     * @param convertibleObject an object that can be converted to a double
     * array or a double scalar.
     */
    public void setVector(Object convertibleObject) {
        this.vector = mapToDoubleVector(convertibleObject);
    }

    /**
     * Tries to convert an object to a double vector.
     *
     * @param object an object.
     * @return a double object.
     */
    public static double[] mapToDoubleVector(Object object) {
        final double[] out;
        if (object instanceof double[]) {
            out = (double[]) object;
        } else if (object instanceof int[]) {
            final int[] in = (int[]) object;
            out = new double[in.length];
            for (int i = 0; i < in.length; i++) {
                out[i] = in[i];
            }
        } else if (object instanceof float[]) {
            final float[] in = (float[]) object;
            out = new double[in.length];
            for (int i = 0; i < in.length; i++) {
                out[i] = in[i];
            }
        } else if (object instanceof short[]) {
            final short[] in = (short[]) object;
            out = new double[in.length];
            for (int i = 0; i < in.length; i++) {
                out[i] = in[i];
            }
        } else if (object instanceof long[]) {
            final long[] in = (long[]) object;
            out = new double[in.length];
            for (int i = 0; i < in.length; i++) {
                out[i] = in[i];
            }
        } else if (object instanceof Double) {
            out = new double[]{(Double) object};
        } else if (object instanceof Integer) {
            out = new double[]{(Integer) object};
        } else if (object instanceof Short) {
            out = new double[]{(Short) object};
        } else if (object instanceof Long) {
            out = new double[]{(Long) object};
        } else if (object instanceof Float) {
            out = new double[]{(Float) object};
        } else {
            throw new ClassCastException("Class "
                    + object.getClass().getCanonicalName()
                    + " is not convertible to double[].");
        }
        return out;
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public String toString() {
        return Arrays.toString(vector);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public int hashCode() {
        int hash = 7;
        hash = 47 * hash + Arrays.hashCode(this.vector);
        return hash;
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final VectorialValue other = (VectorialValue) obj;
        if (!Arrays.equals(this.vector, other.vector)) {
            return false;
        }
        return true;
    }
}
