/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.visualization;
//package de.unibi.citec.fit.dfgvis;

import static de.unibi.citec.tcs.alignment.visualization.Keywords.CODEPOSITION_KEYWORD;
import static de.unibi.citec.tcs.alignment.visualization.Keywords.SOURCE_CODE_FILE_KEYWORD;
//import static de.unibi.citec.fit.dfgvis.Keywords.CODEPOSITION_KEYWORD;
//import static de.unibi.citec.fit.dfgvis.Keywords.SOURCE_CODE_FILE_KEYWORD;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.Value;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import org.apache.commons.lang3.StringEscapeUtils;

/**
 * 
 * 
 * @author ckrueger, bpaassen
 */
public class HTPartStringFromPath extends AbstractHTMLTransformerFromPath {
    /**
     * Holds the line-wise content of the files, given for this keyword.
     */
    private final Map<String, ArrayList<String>> codeFileContent = 
            new HashMap<String, ArrayList<String>>();
    
    /**
     * Attention: when using this constructor, you have to call 
     * setFiles() or setCodeFileContent() manually.
     * @param keyword the keyword, as it should appear in the table.
     */
    public HTPartStringFromPath(final String keyword) {
        super(keyword);
    }
    
    /**
     * Initializes given files or folders. 
     * @param keyword the keyword, as it should appear in the table.
     * @param files_or_folders Paths to source code files or 
     *      a directory/directories containing such files.
     */
    public HTPartStringFromPath(final String keyword, final String[] files_or_folders) {        
        super(keyword, files_or_folders);
    }

    /**
     * Initializes given folder. 
     * @param keyword the keyword, as it should appear in the table.
     * @param folder Paths a directory containing such source code files.
     */
    public HTPartStringFromPath(final String keyword, final String folder) {        
        super(keyword, folder);
    }

    /**
     * Initializes given files or folders. 
     * @param keyword the keyword, as it should appear in the table.
     * @param files_or_folders Source code files or 
     *      a directory/directories containing such files.
     */
    public HTPartStringFromPath(final String keyword, final File[] files_or_folders) {        
        super(keyword, files_or_folders);
    }

    /**
     * Initializes given folder. 
     * @param keyword the keyword, as it should appear in the table.
     * @param folder A directory containing such source code files.
     */
    public HTPartStringFromPath(final String keyword, final File folder) {        
        super(keyword, folder);
    }
    
    /**
     * Returns line-wise content of the specified file. codeFileContent 
     * is gradually added to: when the node given to getHTMLforNode() specifies 
     * a source code file that is not in codeFileContent yet. Therefor 
     * getHTMLforNode() has to be called at least once.
     * Note: equal file name does not include equal file.
     * @param sourceCodeFile File whose content should be returned.
     * @return line-wise content for the specified file.
     */
    public ArrayList<String> getCodeFileContentForKey(final String sourceCodeFile) {
        if (!codeFileContent.containsKey(sourceCodeFile))
            return null;
        
        return codeFileContent.get(sourceCodeFile);
    }
    
    /**
     * Returns codeFileContent, which holds the line-wise content of the 
     * source code files specified by every node. 
     * codeFileContent is gradually added to: when the node given to getHTMLforNode() 
     * specifies a source code file that is not in codeFileContent yet. Therefor 
     * getHTMLforNode() has to be called at least once.
     * @return line-wise content of the files.
     */
    public Map<String, ArrayList<String> > getCodeFileContent() {
        return codeFileContent;
    }
    
    /**
     * Sets codeFileContent, which holds the line-wise content of the 
     * source code files specified by every node.
     * The existing codeFileContent is cleared.
     * @param codeFileContent the new CodeFileContent that will be used.
     */
    public void setCodeFileContent(final Map<String, ArrayList<String> > codeFileContent) {
        this.codeFileContent.clear();
        for (String key : codeFileContent.keySet()) {
            this.codeFileContent.put(key, codeFileContent.get(key));
        }
    }
    
    /**
     * Takes some code positions from the given node and extracts the 
     * corresponding string from a file specified by 'SourceCodeFile' 
     * in the node.
     * @param node a node from some sequence that can be transformed according
     * to this ExternalContent.
     * @return A HTML representation of the extracted content. 
     */
    @Override
    public String getHTMLforNode(final Node node) {
        final String nodeFile = node.getValue(SOURCE_CODE_FILE_KEYWORD).toString();
        
        //searches all assigned directories for nodeFile OR file was directly assigned
        //adds it then to codeFileContent
        if (!codeFileContent.containsKey(nodeFile)) {
            Boolean found = Boolean.FALSE;
            Integer i = 0;
            while (!found && getFiles().containsKey("dir" + i.toString())) {
                for (final File file : getFiles().get("dir" + i.toString()).listFiles()) {
                    if (file.getName().equals(nodeFile)) {
                        codeFileContent.put(nodeFile, Utils.readFile(file));
                        found = Boolean.TRUE;
                        break;
                    }
                }
                i++;
            }
            
            if (!found) {
                throw new RuntimeException("There is no suitable source code file "
                        + "for " + nodeFile + " given");
            }
        }

        final ArrayList<String> sourceCode = codeFileContent.get(nodeFile);
        final Value val = node.getValue(CODEPOSITION_KEYWORD);
        if (val == null) {
            return "<>";
        }
        final String tmpPos = val.toString();
        final String[] codePosition = tmpPos.substring(1, tmpPos.length() - 1).split(", ");

        final int startLine = (int) Double.parseDouble(codePosition[0]) - 1;
        final int startColumn = (int) Double.parseDouble(codePosition[1]) - 1;
        final int endLine = (int) Double.parseDouble(codePosition[2]) - 1;
        final int endColumn = (int) Double.parseDouble(codePosition[3]) - 1;

        //get source code; crop parts with more then five lines 
        if (endLine + 1 > sourceCode.size()) {
            throw new RuntimeException("endLine " + Integer.toString(endLine + 1)
                    + " of node(hash) " + node.hashCode()
                    + " is out of sourceCodeFile's "
                    + nodeFile + " boundaries");
        }
        if (endColumn > sourceCode.get(endLine).length()) {
            throw new RuntimeException("endColumn "
                    + Integer.toString(endColumn + 1) + " of line " + endLine
                    + " of node(hash) " + node.hashCode()
                    + " is out of sourceCodeFile "
                    + nodeFile + "'s boundaries");
        }

        final StringBuilder content = new StringBuilder();
        if (startLine == endLine) {
            if (startColumn == endColumn) {
                content.append(" <> ");
            } else {
                content.append(sourceCode.get(startLine).substring(startColumn, endColumn));
            }
        } else if (endLine < startLine + 5) {
            content.append(sourceCode.get(startLine).substring(startColumn));
            for (int line = startLine + 1; line < endLine; line++) {
                content.append(sourceCode.get(line));
            }
            content.append(sourceCode.get(endLine).substring(0, endColumn));
        } else {
            content.append(sourceCode.get(startLine).substring(startColumn));
            content.append(" [...] ");
            content.append(sourceCode.get(endLine).substring(0, endColumn));
        }
        
        return StringEscapeUtils.escapeHtml4(content.toString());
    }
    
    
    
}
