/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.visualization;

import de.unibi.citec.tcs.alignment.AlignmentPath;
import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.Operation;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.Value;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.file.Files;
import static java.nio.file.StandardCopyOption.REPLACE_EXISTING;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import org.apache.commons.lang3.StringEscapeUtils;

/*
 * (TODO)
 * cell width max
 * tests
 * (/TODO)

FRAGE: image groesse festlegen?
*/

/**
 * Visualizes given alignments in clear and adaptable tables.
 * Stores result to html files.
 * 
 * order of columns: step | total | key1(obj1, obj2, [cost]) | ... | keyn(obj1, obj2,[cost]) | step
 * maximum number of columns = #key*3 + 3 (2x step, total)
 * alignment-table meaning: 
 *                  " --- " : INSERTION or DELETION, so one side is null
 *                  " <> " : side is not null, but for key exists no value
 * @author ckrueger
 */
public class Visualization {

    /**
     * Number of sequences/of rows/columns in the alignment matrix.
     */
    static int NUMBER_OF_OBJECTS;
    
    /**
     * Determines the appearance of the alignment matrix and thus the clickable
     * alignments.
     */
    static boolean upperTriangle;
    
    /**
     * Path of the alignheader.html file as written in the .ini-file.
     */
    static String ALIGNMENT_HEADER_PATH;
    /**
     * Path of the header.html file as written in the .ini-file.
     */
    static String MATRIX_HEADER_PATH;
    /**
     * Path of the format.css file as written in the .ini-file.
     */
    static String FORMAT_CSS_PATH;
    /**
     * Path of the index.html file as written in the .ini-file.
     */
    static String INDEX_HTML_PATH;
    
    /**
     * An error that could be thrown frequently throughout the program.
     */
    final static String ERROR_FILE_PERMISSION = "File or permission error";
    
    /**
     * Loads an .ini-file, which contains different paths, that are needed 
     * throughout the program. This file should be kept local, since the paths
     * on different machines are probably unlike.
     */
    private static void loadIni() {
        final String iniFile = "html_paths.ini";
        try{
            final Properties p = new Properties();
            p.load(new FileInputStream(iniFile));
            final String[] list = {"ALIGNMENT_HEADER_PATH","MATRIX_HEADER_PATH",
                "FORMAT_CSS_PATH","INDEX_HTML_PATH"};
            for (final String str : list) {
                String path = p.getProperty(str);
                final File tmp = new File(path);
                if (!(tmp.exists() && tmp.canRead())) 
                    throw new RuntimeException(ERROR_FILE_PERMISSION + ": " + path);                    
            }
            
            ALIGNMENT_HEADER_PATH = p.getProperty("ALIGNMENT_HEADER_PATH");
            MATRIX_HEADER_PATH = p.getProperty("MATRIX_HEADER_PATH");
            FORMAT_CSS_PATH = p.getProperty("FORMAT_CSS_PATH");
            INDEX_HTML_PATH = p.getProperty("INDEX_HTML_PATH");
            }
        catch (IOException e) {
            throw new RuntimeException("Could not load ini: " + iniFile);
        }        
    }
    
    /**
     * Visualizes given alignments in clear and adaptable tables.
     * Stores result to html files.
     * @param alignments Matrix of alignments.
     * @param usedKeys Keys which should be displayed in the html-output. Put a null 
     *      if you want to use all keys specified in alignment and sequence.
     *      The sequence of keys determines the sequance in the subsequent table.
     * @param datasetName The html-output will be combined in a directory under 
     *      this name. 
     * @param savePath Path were the datasetName-directory will be created.
     * @param programNames Names for the aligned programs.
     * @param extContent Optional; for a description what an external content is
     *  and how it works, see there.
     * @param upperTriangle Determines the appearance of the alignment matrix 
     *  and thus the clickable alignments.
     */
    public static void createHTMLs(final AlignmentPath[][] alignments,
            ArrayList<String> usedKeys, final String datasetName,
            final String savePath, final String[] programNames, 
            ExternalContent[] extContent, final boolean upperTriangle) {
        
        checkCreateHTMLsInput(alignments, usedKeys, datasetName, savePath, 
                programNames, extContent);
        loadIni();
        Visualization.upperTriangle = upperTriangle;
        
        AlignmentPath testAlignment = alignments[0][0];
        both:
        for (int i = 0; i < NUMBER_OF_OBJECTS; i++) {
            for (int j = (Visualization.upperTriangle ? i : 0); j < NUMBER_OF_OBJECTS; j++) {
                if (alignments[i][j] != null) {
                    testAlignment = alignments[i][j];
                    break both;
                }
            }
        }
        if (usedKeys == null || usedKeys.isEmpty()) {
            final Set<String> keys = new HashSet<>();
            keys.addAll(Arrays.asList(testAlignment.getSpecificaton().getKeywords()));
            keys.addAll(Arrays.asList(testAlignment.getSpecificaton()
                    .getNodeSpecification().getKeywords()));
            usedKeys = new ArrayList<>(keys);
        }
                
        if (extContent == null) {
            extContent = new ExternalContent[NUMBER_OF_OBJECTS];
            for (int i = 0; i < NUMBER_OF_OBJECTS; i++) {
                extContent[i] = new ExternalContent();
            }
        }
                
        //computing coloring scheme for every individual keyword
        Map<String, Map<Double, Integer[]> > distColorForKeys = 
                Utils.rgbColorsForKeys(alignments);
        
        //gathering all data for usedKeys from the sequences;
        //searches first horizontally, looking for the left sequence, if every
        //   horizontal alignment in this row is null, it searches vertically looking for the
        //   right sequence, if still every alignment is null, it does not matter,
        //   because the corresponding sequence is never requested
        final Map<Integer, ArrayList<ArrayList<String> > > contentMatrices = 
                new HashMap<>();
        boolean found = Boolean.FALSE;
        for (int i = 0; i < NUMBER_OF_OBJECTS; i++) {
            for (int j = (Visualization.upperTriangle ? i : 0); j < NUMBER_OF_OBJECTS; j++) {
                if (alignments[i][j] != null) {
                        contentMatrices.put(i, buildContentMatrix(alignments[j][i].getRight(), 
                            usedKeys, extContent[i]));
                    found = Boolean.TRUE;
                    break;
                }
            }
            
            if (!found) {
                for (int j = 0; j < (Visualization.upperTriangle ? i + 1 : NUMBER_OF_OBJECTS); j++) {
                    if (alignments[j][i] != null) {
                        contentMatrices.put(i, buildContentMatrix(alignments[j][i].getRight(), 
                            usedKeys, extContent[i]));
                        break;
                    }
                }
            }
            found = Boolean.FALSE;
        }

        //creating alignment tables
        final String form = buildAndFillMatrix(alignments, datasetName, 
                programNames, usedKeys);
        prepareOutput(savePath, datasetName, form);
        
        final StringBuilder text = new StringBuilder();   
        for (final String str : Utils.readFileFromPath(ALIGNMENT_HEADER_PATH)) {
            text.append(str);
            text.append("\n");
        }
        final String alignmentHeader = text.toString();
        
        
        for (int i = 0; i < NUMBER_OF_OBJECTS; i++) {
            for (int j = (Visualization.upperTriangle ? i : 0); j < NUMBER_OF_OBJECTS; j++) {
                if (alignments[i][j] != null) {
                    final String htmlCode = alignmentHeader + visualizeInHTML(
                            alignments[i][j], i, j, contentMatrices.get(i),
                            contentMatrices.get(j), usedKeys, distColorForKeys);
                    
                    saveHTMLOutput(savePath, datasetName, htmlCode, i, j);
                }
            }
        }
    }

    /**
     * Gathers data that should be displayed (defined by usedKeys) and transforms
     * it to html format.
     * @param seq A single sequence.
     * @param usedKeys Keys which will be displayed in the html-output.
     * @param extContent Optional; for a description what an external content is
     *  and how it works, see there.
     * @return A List 
     */
    private static ArrayList<ArrayList<String> > buildContentMatrix(
            final Sequence seq, final ArrayList<String> usedKeys, 
            final ExternalContent extContent) {
        
        final StringBuilder text = new StringBuilder();
        final ArrayList<ArrayList<String> > contentMatrix = 
                new ArrayList<>();
        int counter = 0;
        
        for (final de.unibi.citec.tcs.alignment.sequence.Node node 
                : seq.getNodes()) {
            
            final ArrayList<String> row = new ArrayList<>();
            
            for (final String key : usedKeys) {                
                final String classString = "\n    <td class =\"" + key + "\""
                        + " style=\"display: \" title=\"";
                text.append(classString);
                String content;
                if (!extContent.contains(key)) {
                    Value val = node.getValue(key);
                    if (val == null) content = " <> ";
                    else content = StringEscapeUtils.escapeHtml4(val.toString());
                      
                } else {
                    //extContent
                    content = extContent.getContent(key, counter++);     
                }
                text.append(content);
                text.append("\">");
                text.append(content);
                text.append("</td>");
                row.add(text.toString());
                text.setLength(0);  
            }
            contentMatrix.add(row);
        }
        return contentMatrix;
    }
    
    /**
     * Puts everything together and build a table.
     * @param alignment A single alignment.
     * @param leftIdx The row index of alignment in the alignment matrix.
     * @param rightIdx The column index of alignment in the alignment matrix.
     * @param leftContent The prepared Content for the left sequence of alignment.
     * @param rightContent The prepared Content for the right sequence of alignment.
     * @param usedKeys Keys which should be displayed in the html-output. Put a null 
     *      if you want to use all keys specified in alignment and sequence.
     * @param colorScheme A color scheme for all keys over all alignments.
     * @return A table comparing both sequences of alignment in html format.
     */
    private static String visualizeInHTML(final AlignmentPath alignment, 
            final int leftIdx, final int rightIdx, 
            final ArrayList<ArrayList<String> > leftContent, 
            final ArrayList<ArrayList<String> > rightContent, 
            final ArrayList<String> usedKeys, 
            final Map<String, Map<Double, Integer[]> > colorScheme) {
        
        final ArrayList<String> alignKeywords = new ArrayList<>();
        alignKeywords.addAll(Arrays.asList(alignment.getSpecificaton().getKeywords()));
        
        final StringBuilder text = new StringBuilder();

        text.append("\n <body id=\"alignment\">");
        text.append("\n<h1>Alignment details: ");
        text.append(Utils.formatNumber(leftIdx + 1, 3));
        text.append(" vs. ");
        text.append(Utils.formatNumber(rightIdx + 1, 3));
        text.append("</h1>");
        
        text.append("\n<div id=\"alignment\">");
        text.append("\n<table class=\"table1\" align=\"center\" id=\"tblMain\" ");
        text.append("border=\"1\" style=\"cursor: pointer;\">");
        
        //header
        text.append("\n  <tr>");        
        text.append("\n    <th>Step</th>");
        text.append("\n    <th>Total</th>");
        for (final String key : usedKeys) {
            
            text.append("\n    <th class=\"header\" id=\"");
            text.append(key);
            text.append("\" style=\"display:\" ");
            if (alignKeywords.contains(key)) text.append("colspan=\"3\">");
            else text.append("colspan=\"2\">");
            text.append(key);
            text.append("</th>");
        }
        text.append("\n    <th>Step</th>");
        text.append("\n  </tr>");
        
        final ArrayList<Operation> operations = alignment.getOperations();
        final Map<Double, Integer[]> distColor = Utils.rgbColorsForOperations(operations);
        int leftCounter = 0;
        int rightCounter = 0;
        
        final DecimalFormatSymbols otherSymbols = new DecimalFormatSymbols(Locale.getDefault());
        otherSymbols.setDecimalSeparator('.');
        otherSymbols.setGroupingSeparator(','); 
        final DecimalFormat df = new DecimalFormat( "#,###,###,##0.0000", otherSymbols);
        
        for (int i = 0; i < operations.size(); i++) {
            final Operation op = operations.get(i);
            final ArrayList<String> left;
            final ArrayList<String> right;

            switch (op.getType()) {
                case INSERTION:
//                    case SKIPDELETION:
                    left = null;
                    right = rightContent.get(rightCounter++);
                    break;
                case DELETION: 
//                    case SKIPINSERTION:
                    left = leftContent.get(leftCounter++);
                    right = null;
                    break;
                case REPLACEMENT:
                    left = leftContent.get(leftCounter++);
                    right = rightContent.get(rightCounter++);                        
                    break;
                default:
                    left = null;
                    right = null;                       
            }
            
            //actual table    
            //first step
            text.append("\n  <tr>");
            text.append("\n    <td class=\"Step\">");
            text.append(i + 1);
            text.append("</td>");
            
            //total
            final String scoreString = df.format(op.getWeightedLocalCost());
            final Double score = Double.parseDouble(scoreString);
            text.append("\n    <td style=\"background-color:rgb(");
            text.append(distColor.get(score)[0]);
            text.append(",");
            text.append(distColor.get(score)[1]);
            text.append(",");
            text.append(distColor.get(score)[2]);
            text.append(")\" class=\"OpCosts\">");
            text.append(scoreString);
            text.append("</td>");
            
            //keys
            int keyCounter = 0;
            for (final String key : usedKeys) {
                if (alignKeywords.contains(key)) {
                    final String localScoreString = df.format(
                            op.getComparatorDistances()[alignKeywords.indexOf(key)]);
                    final Double localScore = Double.parseDouble(localScoreString);
                    text.append("\n    <td style=\"display:; background-color:rgb(");
                    text.append(colorScheme.get(key).get(localScore)[0]);
                    text.append(",");
                    text.append(colorScheme.get(key).get(localScore)[1]);
                    text.append(",");
                    text.append(colorScheme.get(key).get(localScore)[2]);
                    text.append(")\" class=\"");
                    text.append(key);
                    text.append("\">");
                    text.append(localScoreString);
                    text.append("</td>");
                }
                
                if (left != null) text.append(left.get(keyCounter));
                else {
                    text.append("\n    <td class =\"");
                    text.append(key);
                    text.append("\" style=\"display:;\">");
                    text.append(" --- ");
                    text.append("</td>");
                }
                
                if (right != null) text.append(right.get(keyCounter));
                else {
                    text.append("\n    <td class =\"");
                    text.append(key);
                    text.append("\" style=\"display:;\">");
                    text.append(" --- ");
                    text.append("</td>");
                }    
                
                keyCounter++;
            }
            
            text.append("\n    <td class=\"Step\">");
            text.append(i + 1);
            text.append("</td>");
            text.append("\n  </tr>");
        }
        text.append("\n</table>");
        text.append("\n</div>");
        
        text.append("\n</body>");
        text.append("\n</html>");
        
        return text.toString();
    }

    /**
     * Checks the input for the visualization of alignments to prevent RuntimeErrors. 
     * Throws RuntimeError if anything is not in order.
     * @param alignments Matrix of alignments.
     * @param usedKeys Keys which should be displayed in the html-output. Put a null 
     *      if you want to use all keys specified in alignment and sequence.
     * @param datasetName The html-output will be combined in a directory under 
     *      this name. 
     * @param savePath Path were the datasetName-directory will be created.
     * @param programNames Names for the aligned programs.
     * @param extContent Optional; for a description what an external content is
     *  and how it works, see there.
     */
    private static void checkCreateHTMLsInput(final AlignmentPath[][] alignments,
            ArrayList<String> usedKeys, final String datasetName,
            final String savePath, final String[] programNames, 
            final ExternalContent[] extContent) {
        
        if (alignments == null) throw new RuntimeException("Alignment matrix is null.");
        AlignmentPath testAlignment = alignments[0][0];
        both:
        for (int i = 0; i < NUMBER_OF_OBJECTS; i++) {
            for (int j = (upperTriangle ? i : 0); j < NUMBER_OF_OBJECTS; j++) {
                if (alignments[i][j] != null) {
                    testAlignment = alignments[i][j];
                    break both;
                }
            }
        }
        if (testAlignment == null) throw new RuntimeException("All alignments are null.");
        
        NUMBER_OF_OBJECTS = alignments.length; 
        if (NUMBER_OF_OBJECTS < 1) throw new RuntimeException("Alignment matrix is empty.");
        if (alignments[0].length != NUMBER_OF_OBJECTS) throw new RuntimeException(
                "Alignment matrix is not square.");
        
        if (savePath == null || savePath.equals("")) throw  new RuntimeException(
            "No savePath given.");
        if (datasetName == null || datasetName.equals("")) throw new RuntimeException(
            "No datasetName given.");
        
        if (usedKeys != null && !usedKeys.isEmpty()) {
            final Set<String> nodeKeywords = new HashSet<>();
            nodeKeywords.addAll(Arrays.asList(testAlignment.getLeft()
                    .getNodeSpecficiation().getKeywords()));
            if (extContent != null) {
                for (final String str : usedKeys) {
                    if (!(extContent[0].contains(str) || nodeKeywords.contains(str)))
                        throw  new RuntimeException("For " + str + " exists no content "
                                + "in extContent or the sequences.");
                }
            } else {
                for (final String str : usedKeys) {
                    if (!nodeKeywords.contains(str))
                        throw  new RuntimeException("For " + str + " exists no content "
                                + "in the sequences.");
                }
            }
        } else {
            final Set<String> keys = new HashSet<>();
            keys.addAll(Arrays.asList(testAlignment.getSpecificaton().getKeywords()));
            keys.addAll(Arrays.asList(testAlignment.getSpecificaton()
                    .getNodeSpecification().getKeywords()));
            
            usedKeys = new ArrayList<>(keys);
      
        }
        
        if (extContent != null) {
            if (extContent.length != NUMBER_OF_OBJECTS) throw new RuntimeException(
                    "Length of extContent does not match " + extContent.length);
            for (final String str : usedKeys) {
                if (extContent[0].contains(str)) {
                    if (extContent[0].getContent(str) == null
                        || extContent[0].getContent(str).size() 
                            != testAlignment.getLeft().getNodes().size())
                    throw new RuntimeException("extContent for keyword " + str 
                            + " is null or does not match sequence length");
                }
            }
        }
        
        if (programNames != null && programNames.length != 0 && programNames.length != alignments[0].length) 
            throw new RuntimeException("There are too few names for the different"
                    + "alignments/programs given.");
    }
    
    /**
     * Builds the form.html with the alignment matrix, the program 
     * names(when given) and the show-column-switches. Applies a coloring 
     * scheme on the matrix.
     * @param alignments Are needed for their scores.
     * @param datasetName Html-output will be combined in a directory called 
     *      datasetName. 
     * @param programNames (optional) Aliases for the sequence numbers.
     * @param usedKeys used for determining the show-column-switches.
     * @return The future form.html.
     */
    private static String buildAndFillMatrix(final AlignmentPath[][] alignments,
            final String datasetName, String[] programNames, 
            final ArrayList<String> usedKeys) {
        
        final StringBuilder text = new StringBuilder();
        final Map<Double, Integer[]> distColor = Utils.rgbColorsForAlignments(alignments);
        for (final String str : Utils.readFileFromPath(MATRIX_HEADER_PATH)) {
            text.append(str);
            text.append("\n");
        }
        final String header = text.toString();
        text.setLength(0);

        text.append("\n<body id=\"form\">");
        text.append("\n<h1>Dissimilarity matrix for data set \"");
        text.append(datasetName);
        text.append("\"</h1>");
        text.append("<div class=\"wrapper\">");
        
        if (programNames == null) programNames = new String[0];
        text.append("\n<div id=\"progNames\">");
        text.append("\n  <ul>");
        for (int i = 0; i < programNames.length; i++) {
            text.append("\n    <li>");
            text.append(Utils.formatNumber(i + 1, 3));
            text.append(" = ");
            text.append(programNames[0]);
            text.append("</li>");
        }
        text.append("\n  </ul>\n</div>");
        
        text.append("\n<div id=\"dismatrix\">");
        text.append("\n<table>");
        text.append("\n  <tr>");
        text.append("\n    <td>L\\R</td>");

        //table header
        for (int i = 1; i < NUMBER_OF_OBJECTS + 1; i++) {
            text.append("\n    <td>");
            text.append(Utils.formatNumber(i, 3));
            text.append("</td>");
        }
        text.append("\n </tr>");

        final DecimalFormatSymbols otherSymbols = new DecimalFormatSymbols(Locale.getDefault());
        otherSymbols.setDecimalSeparator('.');
        otherSymbols.setGroupingSeparator(',');
        final DecimalFormat df = new DecimalFormat( "#,###,###,##0.00", otherSymbols);
        
        double score;
        String scoreString;
        for (int i = 1; i < NUMBER_OF_OBJECTS + 1; i++) {
            text.append("\n  <tr>");
            text.append("\n    <td>");
            text.append(Utils.formatNumber(i, 3));
            text.append("</td>");

            for (int j = 1; j < NUMBER_OF_OBJECTS + 1; j++) {
                if (upperTriangle && i > j) {
                    text.append("<td style=\"visibility: hidden\"></td>");
                    continue;
                }
                
                if (alignments[i - 1][j - 1] == null) {
                    text.append("<td><></td>");
                    continue;
                }
                
                scoreString = df.format(alignments[i - 1][j - 1].getScore());
                score = Double.parseDouble(scoreString);
                text.append("\n    <td style=\"background-color:rgb(");
                text.append(distColor.get(score)[0]);
                text.append(",");
                text.append(distColor.get(score)[1]);
                text.append(",");
                text.append(distColor.get(score)[2]);
                text.append(")\" name=\"link\"><a href=\"Alignments/");
                text.append(Utils.formatNumber(i, 3));
                text.append("_");
                text.append(Utils.formatNumber(j, 3));
                text.append(".html\" target=\"Alignment\">");
                text.append(scoreString);
                text.append("</a></td>");
            }
            
            text.append("\n  </tr>");
        }
        text.append("\n</table>\n</div>");
        
        AlignmentPath testAlignment = alignments[0][0];
        both:
        for (int i = 0; i < NUMBER_OF_OBJECTS; i++) {
            for (int j = (upperTriangle ? i : 0); j < NUMBER_OF_OBJECTS; j++) {
                if (alignments[i][j] != null) {
                    testAlignment = alignments[i][j];
                    break both;
                }
            }
        }
        final AlignmentSpecification spec = testAlignment.getSpecificaton();
        final double[] weights = spec.getWeighting();
        final ArrayList<String> alignKeywords = new ArrayList<>();
        alignKeywords.addAll(Arrays.asList(spec.getKeywords()));
        
        text.append("\n<div id=\"checks\">");
        text.append("\n  <ul>");
        for (final String key : usedKeys) {
            text.append("\n    <li><input type=\"checkbox\" name=\"check\"");
            text.append("checked=\"true\" onchange=\"check();\">");
            text.append(key);
            if (alignKeywords.contains(key)) {
                text.append(" (");
                text.append(df.format(weights[spec.getKeywordIndex(key)]));
                text.append(")");
            }
            text.append("</li>");
        }
        text.append("\n  </ul>\n</div>");
        
        
        text.append("\n</div>\n</body>\n</html>");
        return header + addJavascript(text.toString());
    }
       
    /**
     * Prepends javascript and style code to the html body code.
     * @param htmlCode HTML-Code of the comparison tables (without header).
     * @return Comparison tables with javascript and style code but without header.
     */
    private static String addJavascript(final String htmlCode) {
        String js = ""
                + "<script type=\"text/javascript\">"
                + "\nwindow.onload = function() {"
                + "\n   checks = document.getElementsByName(\"check\");"
                + "\n   var links = document.getElementsByName(\"link\");"
                + "\n   for (var i = 0; i < links.length; i++) {"
                + "\n       links[i].onclick = function() {"
                + "\n           window.setTimeout(check, 80);"
                + "\n       };"
                + "\n   }"
                + "\n};"
                + "\nfunction check() {"
                + "\n   var colCount = 2;"
                + "\n   var tbl = parent.Alignment.document.getElementById(\"tblMain\");" 
                + "\n   //j omitts Step and total" 
                + "\n   for (var j = 2; j < tbl.rows[0].cells.length - 1; j++) {" 
                + "\n       if (checks[j - 2].checked == true && tbl.rows[0].cells[j].style.display != '') {"
                + "\n           tbl.rows[0].cells[j].style.display=\"\";"
                + "\n           for (var n = 0; n < tbl.rows[0].cells[j].colSpan; n++) {" 
                + "\n               for (var i = 1; i < tbl.rows.length; i++) {"
                + "\n                   tbl.rows[i].cells[colCount + n].style.display=\"\";"              
                + "\n               }"
                + "\n           }" 
                + "\n       } else if (checks[j - 2].checked == false && tbl.rows[0].cells[j].style.display != 'none') {"
                + "\n           tbl.rows[0].cells[j].style.display=\"none\";"
                + "\n           for (var n = 0; n < tbl.rows[0].cells[j].colSpan; n++) {" 
                + "\n               for (var i = 1; i < tbl.rows.length; i++) {"
                + "\n                   tbl.rows[i].cells[colCount + n].style.display=\"none\";"              
                + "\n               }"
                + "\n           }" 
                + "\n       }" 
                + "\n" 
                + "\n       colCount = colCount + tbl.rows[0].cells[j].colSpan;" 
                + "\n   }" 
                + "\n}" 
                + "\n</script>";
        
        return js + htmlCode;
    }
    
    /**
     * Creates the output folder, copies index.html, format.css, saves
     * form.html.
     *      format.css.
     * @param datasetName The html-output will be combined in a directory under 
     *      this name. 
     * @param savePath Path were the datasetName-directory will be created.
     * @param form The matrix of alignment scores refering to single alignment 
     *      tables. 
     */
    private static void prepareOutput(final String savePath, 
            final String datasetName, final String form) {       
        final String path = savePath + File.separator + datasetName + File.separator;
        final File dir = new File(path + "Alignments");
        if (!(dir.exists() && dir.isDirectory() && dir.canWrite())) {
            dir.mkdirs();
        }  
        
        try {            
            Files.copy(new File(INDEX_HTML_PATH).toPath(), new File(path + "index.html")
                .toPath(), REPLACE_EXISTING);
            Files.copy(new File(FORMAT_CSS_PATH).toPath(), new File(path + "format.css")
                .toPath(), REPLACE_EXISTING);
        } catch (IOException ex) {
            throw new RuntimeException("Copying index.html and format.css failed."
                    + " IOexception: " + ex.getMessage());
        }
        
        Utils.writeFile(path + "form.html", form);
    }
    
    /**
     * Writes one alignment table to the given path.
     * @param datasetName The html-output will be combined in a directory under 
     *      this name. 
     * @param savePath Path were the datasetName-directory will be created.
     * @param htmlCode One alignment table in html format.
     */
    private static void saveHTMLOutput(final String savePath, 
            final String datasetName, final String htmlCode, final int i, final int j) {
        
        final String path = savePath + File.separator + datasetName + File.separator;       
        
        if (htmlCode != null) {
            final String file = path + "Alignments" + File.separator
                    + Utils.formatNumber(i + 1, 3) + "_" + Utils.formatNumber(j + 1, 3)
                    + ".html";
            Utils.writeFile(file, htmlCode);
        }
    }    
}
