/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.Comparator;
import de.unibi.citec.tcs.alignment.comparators.DTWKernelComparator;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import java.util.Arrays;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author gzentgra
 */
public class DynamicTimeWarpingAlgorithmTest {

	public DynamicTimeWarpingAlgorithmTest() {
	}

	@BeforeClass
	public static void setUpClass() {
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of calculateAlignment method, of class DynamicTimeWarpingAlgorithm.
	 */
	@Test
	public void testCalculateAlignment() {
		/*
		 * test DTW alignment and derivativs:
		 * compare java calculations with precalculated results in matlab
		 */

        // nodes contain 3 keywords, each are vectors since so far I only have
		// that one comparator
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[3];
		keywordSpecs[0] = new VectorialKeywordSpecification(1, "vec1");
		keywordSpecs[1] = new VectorialKeywordSpecification(2, "vec2");
		keywordSpecs[2] = new VectorialKeywordSpecification(3, "vec3");

		// derive the node specification from that.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);

		// create sqeuences (same as in matlab)
		final Sequence x = new Sequence(nodeSpec);

		Node node = new Node(x);
		node.setValue("vec1", new VectorialValue(new double[]{0}));
		node.setValue("vec2", new VectorialValue(new double[]{1, 0}));
		node.setValue("vec3", new VectorialValue(new double[]{-2, 2, 1}));
		x.getNodes().add(node);
		node = new Node(x);
		node.setValue("vec1", new VectorialValue(new double[]{1}));
		node.setValue("vec2", new VectorialValue(new double[]{0, 1}));
		node.setValue("vec3", new VectorialValue(new double[]{0, 1, 0}));
		x.getNodes().add(node);
		node = new Node(x);
		node.setValue("vec1", new VectorialValue(new double[]{0}));
		node.setValue("vec2", new VectorialValue(new double[]{1, 1}));
		node.setValue("vec3", new VectorialValue(new double[]{0, 0, -1}));
		x.getNodes().add(node);

		final Sequence y = new Sequence(nodeSpec);
		node = new Node(y);
		node.setValue("vec1", new VectorialValue(new double[]{-1}));
		node.setValue("vec2", new VectorialValue(new double[]{1, 0}));
		node.setValue("vec3", new VectorialValue(new double[]{1, 1, 1}));
		y.getNodes().add(node);
		node = new Node(y);
		node.setValue("vec1", new VectorialValue(new double[]{-1}));
		node.setValue("vec2", new VectorialValue(new double[]{0, 1}));
		node.setValue("vec3", new VectorialValue(new double[]{0, 0, 0}));
		y.getNodes().add(node);

		// specify alignment
		final String[] keywords = new String[]{"vec1", "vec2", "vec3"};
		final DTWKernelComparator[] comparators = new DTWKernelComparator[3];
		comparators[0] = new DTWKernelComparator(new double[]{1}, 1);
		comparators[1] = new DTWKernelComparator(new double[]{0.2, 0.8}, 1);
		comparators[2] = new DTWKernelComparator(new double[]{1.0 / 3, 1.0 / 3, 1.0 / 3}, 1);
		final double[] weights = new double[]{0.2, 0.5, 0.3};

		final AlignmentSpecification alignmentSpec
				= new AlignmentSpecification(nodeSpec, keywords, comparators, weights);

//------------------------------- calc alignment -------------------------------        
		final DynamicTimeWarpingAlgorithm algo = new DynamicTimeWarpingAlgorithm(alignmentSpec);

		SoftDTWModel result = algo.calculateAlignment(x, y);

//-------------------- check results / calc derivative ------------------------- 
		expResult expRes = new expResult();

		// test accumulated cost mat
		assertTrue(expRes.testDTWCostMat(result.getSimilarityMatrix(), expRes.getAccuracy()));

		// test local derivation
		double[] localDev = new double[6];
		localDev[0] = result.calculateParameterDerivative(comparators[0], "vec1")[0];
		localDev[1] = result.calculateParameterDerivative(comparators[1], "vec2")[0];
		localDev[2] = result.calculateParameterDerivative(comparators[1], "vec2")[1];
		localDev[3] = result.calculateParameterDerivative(comparators[2], "vec3")[0];
		localDev[4] = result.calculateParameterDerivative(comparators[2], "vec3")[1];
		localDev[5] = result.calculateParameterDerivative(comparators[2], "vec3")[2];
        // TODO: change matlab alg and this alg!!
		//assertTrue( expRes.testDTWLocalDerivation(localDev, expRes.getAccuracy()) );

		// test weight derivation
		assertTrue(expRes.testDTWWeightDerivation(result.calculateWeightDerivative(), expRes.getAccuracy()));
	}

//------------------------------------------------------------------------------
	// harcodes expected result values and implements easy test functions
	private static class expResult {

		// number correct of decimal places in expResult values
		private final int accuracy;

		private final double[][] DTW_Cost_Mat;
		private final double[] DTW_Local_Dev;
		private final double[] DTW_Weight_Dev;

		// hard code expected values calcualted with matlab
		public expResult() {
			this.accuracy = 4;
			this.DTW_Cost_Mat = new double[4][3];
			this.DTW_Cost_Mat[0][0] = 1;
			this.DTW_Cost_Mat[1][0] = 0;
			this.DTW_Cost_Mat[2][0] = 0;
			this.DTW_Cost_Mat[3][0] = 0;
			this.DTW_Cost_Mat[0][1] = 0;
			this.DTW_Cost_Mat[1][1] = 0.5128;
			this.DTW_Cost_Mat[2][1] = 0.1922;
			this.DTW_Cost_Mat[3][1] = 0.0760;
			this.DTW_Cost_Mat[0][2] = 0;
			this.DTW_Cost_Mat[1][2] = 0.1671;
			this.DTW_Cost_Mat[2][2] = 0.5588;
			this.DTW_Cost_Mat[3][2] = 0.5839;

			this.DTW_Local_Dev = new double[]{0.3289, 0.1751, 0.4252,
				0.1422, 0.4072, 0.3202};

			this.DTW_Weight_Dev = new double[]{2.5144, 4.4394, 3.1436};
		}

		public int getAccuracy() {
			return this.accuracy;
		}

		// testMatAcc: number correct of decimal places in testMat values
		public boolean testDTWCostMat(double[][] testMat, int testMatAcc) {
			if (testMat.length != this.DTW_Cost_Mat.length
					|| testMat[0].length != this.DTW_Cost_Mat[0].length) {
				return false;
			}
			// get min accuracy
			int acc = Math.min(this.accuracy, testMatAcc);
			for (int i = 0; i < testMat.length; i++) {
				for (int j = 0; j < testMat[0].length; j++) {
                    // if difference between test and expected value is to high
					// return false
					if (Math.abs(testMat[i][j] - this.DTW_Cost_Mat[i][j]) > Math.pow(10, -acc)) {
						return false;
					}
				}
			}
			return true;
		}

		// testMatAcc: number correct of decimal places in testDev values
		public boolean testDTWLocalDerivation(double[] testDev, int testMatAcc) {
			// get min accuracy
			int acc = Math.min(this.accuracy, testMatAcc);
			return compareDVec(testDev, this.DTW_Local_Dev, acc);
		}

		// testMatAcc: number correct of decimal places in testDev values
		public boolean testDTWWeightDerivation(double[] testDev, int testMatAcc) {
			// get min accuracy
			int acc = Math.min(this.accuracy, testMatAcc);
			return compareDVec(testDev, this.DTW_Weight_Dev, acc);
		}

		// piecewise compares elements of given double vectors for given acc
		private boolean compareDVec(double[] v1, double[] v2, int acc) {
			if (v1.length != v2.length) {
				return false;
			}
			for (int i = 0; i < v1.length; i++) {
				// if difference between values is to high return false
				if (Math.abs(v1[i] - v2[i]) > Math.pow(10, -acc)) {
					return false;
				}
			}
			return true;
		}
	}
}
