/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import java.util.Random;
import static junit.framework.Assert.assertEquals;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class SoftPathModelTest extends TestCase {

	public SoftPathModelTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of calculateSoftAlignment method, of class SoftAlignmentAlgorithm.
	 *
	 * This speed test is too slow for practical purposes and will be
	 * uncommented on demand.
	 */
	public void testPerformance() {

		final boolean showResults = false;
		//test for different lengths, number of keywords and alphabet sizes
		final int max_length = 50;
		final int length_steps = 10;
		final int max_num_keywords = 5;
		final int max_alpha_size = 25;
		final int alpha_steps = 5;
		final Random rand = new Random();
		for (int A = alpha_steps; A <= max_alpha_size; A += alpha_steps) {
			if (showResults) {
				System.out.println("Number of Symbols: " + A);
			}
			final StringBuilder alphaString = new StringBuilder();
			for (int a = 0; a < A; a++) {
				alphaString.append((char) (97 + a));
				alphaString.append('|');
			}
			final Alphabet alpha = new Alphabet(alphaString.toString());
			for (int K = 1; K <= max_num_keywords; K++) {
				if (showResults) {
					System.out.println("Number of keywords: " + K);
				}
				final KeywordSpecification[] keywordSpecs = new KeywordSpecification[K];
				for (int k = 0; k < K; k++) {
					keywordSpecs[k] = new SymbolicKeywordSpecification(
							alpha, "keyword" + k);
				}
				final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
				final AlignmentSpecification alignSpec = new AlignmentSpecification(nodeSpec);
				final SoftAlignmentFullAlgorithm algo
						= new SoftAlignmentFullAlgorithm(alignSpec);
				algo.setBeta(10);
				for (int L = length_steps; L <= max_length; L += length_steps) {
					if (showResults) {
						System.out.println("Sequence length: " + L);
					}
					//prepare two sequences of random symbols.
					final Sequence a = new Sequence(nodeSpec);
					final Sequence b = new Sequence(nodeSpec);
					for (int i = 0; i < L; i++) {
						final Node aNode = new Node(a);
						for (int k = 0; k < K; k++) {
							final String symbol = Character.toString(
									(char) (97 + rand.nextInt(A)));
							aNode.setValue(k, new SymbolicValue(alpha, symbol));
						}
						a.getNodes().add(aNode);
						final Node bNode = new Node(a);
						for (int k = 0; k < K; k++) {
							final String symbol = Character.toString(
									(char) (97 + rand.nextInt(A)));
							bNode.setValue(k, new SymbolicValue(alpha, symbol));
						}
						b.getNodes().add(bNode);
					}
					//now we have our sequences and can do the alignment.
					final SoftPathModel instance = algo.calculateAlignment(a, b);
					final long start = System.nanoTime();
					final ReplacementComparator comparator = (ReplacementComparator) instance.getSpecificaton().getComparator(0);
					final double[][] derivative = instance.calculateParameterDerivative(
							comparator, "keyword0");
					double[] calculateWeightDerivative = instance.calculateWeightDerivative();
					final long time = System.nanoTime() - start;
					final long milliseconds = time / (long) 1E6;
					if (showResults) {
						System.out.println(milliseconds + "ms");
					}
				}
			}
		}
	}

	/**
	 * Test of calculateDerivative method, of class SoftPathModel.
	 */
	public void testCalculateDerivative() {
		/*
		 * We use a toy example here with an alignment of the strings
		 *
		 * ab
		 *
		 * and
		 *
		 * aab
		 *
		 * where a->a and b->b replacement are used as well as an a insertion,
		 * but nothing else.
		 */
		final Sequence a = new Sequence("ab");
		final Sequence b = new Sequence("aab");

		//use soft alignment with high beta
		final SoftAlignmentFullAlgorithm algo = new SoftAlignmentFullAlgorithm(new AlignmentSpecification(a));
		algo.setBeta(15);
		final SoftPathModel instance = algo.calculateAlignment(a, b);
		//calculate derivative
		final ReplacementComparator comparator = (ReplacementComparator) instance.getSpecificaton().getComparator(0);
		final double[][] derivative = instance.calculateParameterDerivative(
				comparator, Sequence.DEFAULTKEYWORD);

		//replacement derivative should be 1.
		assertEquals(1, derivative[0][0], 1E-4);
		assertEquals(1, derivative[1][1], 1E-4);
		//deletion deriative should be 0
		assertEquals(0, derivative[0][2], 1E-4);
		assertEquals(0, derivative[1][2], 1E-4);
		//we insert an a, but not a b
		assertEquals(1, derivative[2][0], 1E-4);
		assertEquals(0, derivative[2][1], 1E-4);
	}

	/**
	 * Test of calculateWeightDerivative method, of class SoftPathModel.
	 */
	public void testCalculateWeightDerivative() {
		/*
		 * we use a toy example here, where one feature does not differ at all
		 * while another one does, namely a java program snippet with variable
		 * declarations with different classes.
		 */
		//specify the keywords.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[2];
		//first the node type.
		final Alphabet nodeTypes = new Alphabet(new String[]{"double", "int"});
		keywordSpecs[0] = new SymbolicKeywordSpecification(nodeTypes, "class");
		//then the name
		keywordSpecs[1] = new StringKeywordSpecification("name");
		//derive the node specification from that.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//than create the sequences (one double and one int based)
		final Sequence a = new Sequence(nodeSpec);
		Node node = new Node(a);
		node.setValue("class", new SymbolicValue(nodeTypes, "int"));
		node.setValue("name", new StringValue("i"));
		a.getNodes().add(node);
		node = new Node(a);
		node.setValue("class", new SymbolicValue(nodeTypes, "int"));
		node.setValue("name", new StringValue("j"));
		a.getNodes().add(node);

		final Sequence b = new Sequence(nodeSpec);
		node = new Node(b);
		node.setValue("class", new SymbolicValue(nodeTypes, "double"));
		node.setValue("name", new StringValue("i"));
		b.getNodes().add(node);
		node = new Node(b);
		node.setValue("class", new SymbolicValue(nodeTypes, "double"));
		node.setValue("name", new StringValue("j"));
		b.getNodes().add(node);

		//calculate a soft alignment first.
		final AlignmentSpecification alignSpec = new AlignmentSpecification(nodeSpec);

		final SoftAlignmentFullAlgorithm algo = new SoftAlignmentFullAlgorithm(alignSpec);
		algo.setBeta(15);
		final SoftPathModel instance = algo.calculateAlignment(a, b);

		//and calculate a strict alignment for comparison.
		final StrictAlignmentFullAlgorithm strict = new StrictAlignmentFullAlgorithm(alignSpec);
		final AlignmentPath alignment = strict.calculateAlignment(a, b);
		System.out.println(alignment.toString());

		//check the derivative.
		final double[] expected = alignment.calculateWeightDerivative();
		final double[] actual = instance.calculateWeightDerivative();

		assertEquals(2, expected.length);
		assertEquals(expected.length, actual.length);

		for (int k = 0; k < expected.length; k++) {
			assertEquals(expected[k], actual[k], 0);
		}
	}

}
