/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.ParallelProcessingEngine;
import de.unibi.citec.tcs.alignment.StrictAlignmentScoreAlgorithm;
import de.unibi.citec.tcs.alignment.comparators.NCDComparator.CompressorType;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import java.util.Random;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class NCDComparatorTest extends TestCase {

	public NCDComparatorTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of compare method, of class NCDComparator.
	 */
	public void testCompare() {
		final String testStr = "This is a test text of some length.";
		final String testStr2 = "While we now have something entirely different";
		final NCDComparator instance = new NCDComparator();
		for (final CompressorType compressor : CompressorType.values()) {
			instance.setCompressor(compressor);
			double result = instance.compare(
					new StringValue(testStr),
					new StringValue(testStr));
			assertTrue("Result was expected to be small but was " + result, result < 0.3);
			result = instance.compare(new StringValue(testStr),
					new StringValue(testStr2));
			assertTrue("Result was expected to be large but was " + result, result > 0.5);
		}
	}

	public void testPerformance() {
		final int seqNum = 15;
		final int seqLength = 15;
		final int strLength = 3000;

		//create test sequences.
		final KeywordSpecification[] keywordSpecs = {new StringKeywordSpecification("test")};
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		final Sequence[] sequences = new Sequence[seqNum];
		final Random rand = new Random();
		for (int s = 0; s < seqNum; s++) {
			sequences[s] = new Sequence(nodeSpec);
			for (int i = 0; i < seqLength; i++) {
				final Node node = new Node(sequences[s]);
				sequences[s].getNodes().add(node);
				final StringBuilder builder = new StringBuilder();
				for (int c = 0; c < strLength; c++) {
					final char ch = (char) (32 + rand.nextInt(90));
					builder.append(ch);
				}
				node.setValue(0, new StringValue(builder.toString()));
			}
		}

		//use NCD within alignment.
		final NCDComparator instance = new NCDComparator();
		instance.setCompressor(CompressorType.LZ4);
		final AlignmentSpecification alignSpec = new AlignmentSpecification(
				nodeSpec, new String[]{"test"},
				new Comparator[]{instance});

		final StrictAlignmentScoreAlgorithm algo
				= new StrictAlignmentScoreAlgorithm(alignSpec);

		final ParallelProcessingEngine<Double> parallelProc
				= new ParallelProcessingEngine<>(algo, sequences);
		parallelProc.setNumberOfThreads(32);
		parallelProc.setVerbose(false);
		final long start = System.nanoTime();
		parallelProc.calculate();
		final long time = (System.nanoTime() - start) / (long) 1E6;
		System.out.println("NCD for " + seqNum + " sequences of " + seqLength
				+ " nodes each with in turn string content of length " + strLength
				+ " each needed " + time + " milliseconds.");
	}

}
